/***************************************************************************
                          curvefittingmanager.cpp  -  description
                             -------------------
    begin                : Tue Nov 12 19:20:03 CEST 2019
    copyright            : (C) 2019 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "curvefittingmanager.h"
#include "3rdparty/alglib/src/interpolation.h"
#include <QDebug>

CurveFittingManager::CurveFittingManager()
    : maxits(0), epsx(0.0), diffstep(0.0001)
{
}

QList<GenericCurve *> & CurveFittingManager::curvesList()
{
    static QList<GenericCurve *> curves = QList<GenericCurve*>();
    return curves;
}

void CurveFittingManager::setControls(alglib::ae_int_t m, double e, double d)
{
    maxits = m;
    epsx = e;
    diffstep = d;
}

void CurveFittingManager::setScanData(const Scan &_scan)
{
    dataRange = _scan;
    dataRange.pDataHkl().clear();
}

void CurveFittingManager::setCurves(const QList<GenericCurve *> &_clist)
{
    curvesList() = _clist;
}

void CurveFittingManager::setValues(const QVector<double> &_initVal, const QVector<double> &_loLimits, const QVector<double> &_upLimits)
{
    initValues   = _initVal;
    lowerLimits  = _loLimits;
    upperLimits  = _upLimits;
}

QVector<double> CurveFittingManager::fit()
{
    int params = 0;

    for (int i = 0; i < curvesList().size(); ++i) {
        params += curvesList().at(i)->nParameters();
    }

    qDebug() << QString("CurveFittingManager::fit(): Fitting %1 curves to scan.").arg(curvesList().size());

    if ((params != initValues.size()) || (params != lowerLimits.size()) || (params != upperLimits.size())) {
        qDebug() << QString("CurveFittingManager::fit(): Wrong parameter count (expected: %1 received: %2). Exiting.").arg(params).arg(initValues.size());
        return QVector<double>();
    }

    iteration = 1;

    alglib::real_2d_array xdata;
    alglib::real_1d_array ydata;
    alglib::real_1d_array c;
    alglib::real_1d_array bndl;
    alglib::real_1d_array bndu;

    c.setlength(params);
    bndl.setlength(params);
    bndu.setlength(params);

    xdata.setcontent(dataRange.pDataAngle().size(), 1, dataRange.pDataAngle().data());
    ydata.setcontent(dataRange.pDataIntensity().size(), dataRange.pDataIntensity().data());
    c.setcontent(initValues.size(), initValues.data());
    bndl.setcontent(lowerLimits.size(), lowerLimits.data());
    bndu.setcontent(upperLimits.size(), upperLimits.data());

    alglib::ae_int_t info;
    alglib::lsfitstate state;
    alglib::lsfitreport rep;

    alglib::lsfitcreatef(xdata, ydata, c, diffstep, state);
    alglib::lsfitsetbc(state, bndl, bndu);
    alglib::lsfitsetcond(state, epsx, maxits);
    alglib::lsfitfit(state, fitCurve, stepOutput);
    alglib::lsfitresults(state, info, c, rep);

    fitReport = generateReportString(curvesList(), info, rep);

    if (int(info) < 0) {
        return QVector<double>();
    }

    for (int i = 0; i < initValues.size(); ++i) {
        initValues[i] = c[i];
    }

    return initValues;
}

QVector<double> CurveFittingManager::getXvalues(double stepSize)
{
    QVector<double> xcalc;

    // if stepsize < 0, use the stepsize of the data range. Else use the
    // specified stepsize for the calculated curves
    if (stepSize < 0.0) {
        xcalc = dataRange.pDataAngle();
    } else {
        double curang = dataRange.pDataAngle().first();
        double ss = qFuzzyIsNull(stepSize) ? 0.0001 : stepSize;

        xcalc = QVector<double>((dataRange.pDataAngle().last() - curang)/ss, 0.0);

        for (int i = 0; i < xcalc.size(); ++i) {
            xcalc[i] = curang;
            curang += ss;
        }
    }

    return xcalc;
}

void CurveFittingManager::fitCurve(const alglib::real_1d_array &c, const alglib::real_1d_array &x, double &func, void *)
{
    func = curvesList()[0]->fit(c, x);

    for (int i = 1; i < curvesList().size(); ++i) {
        func += curvesList()[i]->fit(c, x);
    }
}

void CurveFittingManager::stepOutput(const alglib::real_1d_array &, double, void *)
{
    /* todo */
}

QList<Scan> CurveFittingManager::calculateCurve(double stepSize, const QString &sumName, const QUuid &sumUid)
{
    QVector<double> x = getXvalues(stepSize);

    alglib::real_1d_array c;
    c.setcontent(initValues.size(), initValues.data());

    QList<Scan> curves;
    QVector<double> sumY(x.size(), 0.0);

    for (int i = 0; i < curvesList().size(); ++i) {
        QVector<double> y(x.size(), 0.0);
        curvesList()[i]->compute(c, x, y, sumY);

        Scan sc = dataRange.clone();
        sc.setUid(curvesList().at(i)->uid());
        sc.setColor(QColor());
        sc.setName(curvesList().at(i)->displayName());
        sc.setTypes(Scan::XY | Scan::SYNTHETIC | Scan::TEMPORARY);
        sc.setXoffset(0.0);
        sc.setDataAng(x);
        sc.setDataInt(y);

        curves.append(sc);
    }

    Scan sum = dataRange.clone();
    sum.setUid(sumUid);
    sum.setColor(QColor());
    sum.setName(QString("#S %1").arg(sumName));
    sum.setTypes(Scan::XY | Scan::SYNTHETIC | Scan::TEMPORARY);
    sum.setXoffset(0.0);
    sum.setDataAng(x);
    sum.setDataInt(sumY);

    curves.append(sum);
    return curves;
}

double CurveFittingManager::getArea(int n, QUuid &uid, bool &ok)
{
    if (n >= curvesList().size()) {
        ok = false;
        return -1.0;
    }

    ok = true;
    uid = curvesList().at(n)->uid();
    return curvesList().at(n)->area();
}

QMap<QString, QVariant> CurveFittingManager::generateReportString(const QList<GenericCurve *> &, const alglib::ae_int_t &i, const alglib::lsfitreport &r)
{
    QMap<QString, QVariant> rep;

    int errorCode = int(i);
    QString errorString;

    switch (errorCode) {
    case -8:
        errorString = QString("optimizer detected NAN/INF");
        break;
    case -7:
        errorString = QString("gradient verification failed");
        break;
    case -3:
        errorString = QString("inconsistent restraints");
        break;
    case 2:
        errorString = QString("converged");
        break;
    case 5:
        errorString = QString("MaxIts steps reached");
        break;
    case 7:
        errorString = QString("stopping conditions are too stringent");
        break;
    default:
        errorString = QString("unknown error code").arg(i);
    }

    rep["ExitCode"] = errorCode;
    rep["ExitString"] = errorString;
    rep["R2"] = r.r2;

    return rep;
}
