/***************************************************************************
                          bgmndiaimport.cpp  -  description
                             -------------------
    begin                : Thu Jan 26, 2012
    copyright            : (C) 2012 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "bgmndiaimport.h"
#include <stdio.h>
#include <QtConcurrent>

BgmnDiaImport::BgmnDiaImport(QObject *parent) :
    GenericImport(parent)
{
    extens << "dia";
    descr = QLatin1String("BGMN refined profile");

    _addBackground = true;
}

/************* PUBLIC FUNCTIONS *************/

/* - If the load function returns -1, don't use any of the
     get* functions. They are bound to break!

   - Always make sure that the load function was called at least once
     before using the get* functions. Otherwise you'll get an error message.
*/

bool BgmnDiaImport::isSupported(const QByteArray &ba)
{
    if (ba.contains(QByteArray("TITEL="))) {
        return true;
    }

    return false;
}

int BgmnDiaImport::load( const QString &file, QVector<Scan> &scanHeap, bool minimal)
{
    Q_UNUSED(minimal)

    _multiScans = 1;

    if (!parseBinaryData(_data, BgmnFileIO::readBinaryFileLines(file))) {
        qDebug() << QString("BgmnDiaImport::load(): Could not read file %1").arg(file);
        return -1;
    }

    int ll;
    int tl;

    if (checkSize(ll, tl) <= 0) return -1;
    return createScans(scanHeap, file);
}


int BgmnDiaImport::loadAndCheck( const QString &file, QVector<Scan> &scanHeap, int &ll, int &tl, bool minimal)
{
    Q_UNUSED(minimal)

    _multiScans = 1;

    if (!parseBinaryData(_data, BgmnFileIO::readBinaryFileLines(file))) {
        qDebug() << QString("BgmnDiaImport::load(): Could not read file %1").arg(file);
        ll = 0;
        tl = std::numeric_limits<int>().max();
        return -1;
    }

    if (checkSize(ll, tl) <= 0) return -1;
    return createScans(scanHeap, file);
}
/************* PRIVATE FUNCTIONS *************/

int BgmnDiaImport::checkSize(int &ll, int &tl)
{
    if (_data.size() < 3) {
        qDebug() << QString("BgmnDiaImport::checkSize(): Incomplete file, found %1 data points").arg(_data.size());
        ll = 0;
        tl = std::numeric_limits<int>().max();
        return -1;
    }

    QRegularExpression rx("M=(\\d+)(?= PEAKZAHL)");
    QRegularExpressionMatch rm = rx.match(_header);

    if (!rm.hasMatch()) {
        qDebug() << QString("BgmnDiaImport::checkSize(): Could not parse header line: %1").arg(QString(_header));
        ll = 0;
        tl = std::numeric_limits<int>().max();
        return -1;
    }

    int m = rm.captured(1).toInt();

    ll = _data.size() * _multiScans;
    tl = m;

    if (m == _data.size() * _multiScans) {
        return 1;
    }

    qDebug() << QString("BgmnDiaImport::checkSize(): Incomplete file, found %1 data points, expected %2").arg(_data.size() * _multiScans).arg(m);
    return -1;
}

int BgmnDiaImport::createScans(QVector<Scan> &scanHeap, const QString &sourceFile)
{
        QStringList struc = getStructureNames(_header);
        double wavelength = getWaveLength(_header);
        double diffMax = 0.0;

        int nScans  = _data.first().size() - 1;

        if (nScans < 3) {
            qDebug() << QString("BgmnDiaImport::getScans(): The file only contains %1 scans, not a valid DIA file.").arg(nScans);
            return -1;
        }

        scanHeap.clear();

        // only intensity vectors must be initialized. Angle vectors will be copied later.
        // add one more for the difference curve
        for (int i = 0; i <= nScans; ++i) {
            scanHeap.append(Scan());
            scanHeap.last().pDataIntensity() = QVector<double>(_data.size(), 0.0);
            scanHeap.last().setSourceFileName(sourceFile);
        }

        scanHeap[0].pDataAngle() = QVector<double>(_data.size(), 0.0);

        // layout of the scanHeap:
        // [0] = iObs
        // [1] = iCalc
        // [2] = iDiff
        // [3] = iBkgr
        // [4...n] = strucs

        for (int pos = 0; pos < _data.size(); ++pos) {
            const QVector<double> *line = &(_data.at(pos));

            scanHeap[0].pDataAngle()[pos] = line->at(0);

            for (int i = 0; i < line->size(); ++i) {
                if (i < 2)  scanHeap[i].pDataIntensity()[pos] = line->at(i + 1);
                if (i == 2) scanHeap[i].pDataIntensity()[pos] = line->at(1) - line->at(2);
                if (i == 3) scanHeap[i].pDataIntensity()[pos] = line->at(i);
                if (i > 3)  scanHeap[i].pDataIntensity()[pos] = line->at(i) + line->at(3);
            }

            diffMax = qMax(diffMax, scanHeap.at(2).pDataIntensity().at(pos));
        }

        for (int i = 1; i < scanHeap.size(); ++i) {
            scanHeap[i].pDataAngle() = scanHeap[0].pDataAngle();
        }

        // add additional info to the scans
        for (int i = 0; i < scanHeap.size(); ++i) {
            scanHeap[i].setWaveLength(wavelength);

            // copy the angle vector from the first scan to the others
            if (i > 0) {
                scanHeap[i].pDataAngle() = scanHeap[0].pDataAngle();
            }

            if (i == 0) {
                scanHeap[i].setName("I observed");
                scanHeap[i].setColor(QColor("#000000"));
                scanHeap[i].setTypes(Scan::XY | Scan::MEASURED | Scan::REFINED);
            }

            if (i == 1) {
                scanHeap[i].setName("I calculated");
                scanHeap[i].setColor(QColor("#ff0000"));
                scanHeap[i].setTypes(Scan::XY | Scan::REFINED);
            }

            if (i == 2) {
                scanHeap[i].setName("I difference");
                scanHeap[i].setColor(QColor("#bbbbbb"));
                scanHeap[i].setTypes(Scan::XY | Scan::REFINED | Scan::DIFF);
            }

            if (i == 3) {
                scanHeap[i].setName("Background");
                scanHeap[i].setColor(QColor("#0000ff"));
                scanHeap[i].setTypes(Scan::XY | Scan::REFINED | Scan::BACKGROUND);
            }

            if (i > 3) {
                if ((i - 4) < struc.size()) {
                    scanHeap[i].setName(struc.at(i - 4));
                    scanHeap[i].setTypes(Scan::XY | Scan::REFINED | Scan::PHASE);
                } else {
                    scanHeap[i].setName("Unknown Phase");
                    scanHeap[i].setTypes(Scan::XY | Scan::REFINED | Scan::PHASE);
                }
                scanHeap[i].setColor(QColor());
            }

        }

        if (scanHeap.size() > 3) {
            scanHeap[2].setYoffset(-1.2 * diffMax);
        }

        return scanHeap.size();
}

QStringList BgmnDiaImport::getStructureNames(const QByteArray &line)
{
    QRegularExpression rxStruc("STRUC\\[\\d+\\]=(\\S+)");
    QRegularExpressionMatch rm = rxStruc.match(line);
    QStringList struc;

    while (rm.hasMatch()) {
        struc.append(rm.captured(1));
        rm = rxStruc.match(_header, rm.capturedEnd(1) + 1);
    }

    return struc;
}

double BgmnDiaImport::getWaveLength(const QByteArray &line)
{
    double wl = 1.54056;

    QRegularExpression rx("L0=(-?\\d+\\.?\\d*)");
    QRegularExpressionMatch rm = rx.match(line);

    if (rm.hasMatch()) {
        wl = rm.captured(1).toDouble();
        if (!qFuzzyIsNull(wl)) {
            wl = 10.0 / qAbs(wl);
        } else {
            qDebug() << QString("BgmnDiaImport::getWaveLength(): L0 is not valid, returning fallback value of 1.54056.");
        }
    } else {
        qDebug() << QString("BgmnDiaImport::getWaveLength(): L0 not found, returning fallback value of 1.54056.");
    }

    return wl;
}

bool BgmnDiaImport::parseBinaryData(QList<QVector<double> > &data, const QList<QByteArray> &lines)
{
    if (lines.size() < 2) return false;

    _header = lines.first();
    _multiScans = countMultiScans(lines);

    QList<QByteArray>::const_iterator it = lines.constBegin();
    ++it; // skip title line

    QFuture<QVector<double> > f = QtConcurrent::mapped(it, lines.constEnd(), parseLine);
    f.waitForFinished();
    data = f.results();

    if (_multiScans > 1) data = mergeMultiScans(data, _multiScans);

    return (data.size() > 0);
}

QVector<double> BgmnDiaImport::parseLine(const QByteArray &line)
{
    QList<QByteArray> fields = line.simplified().split(' ');
    QVector<double> data(fields.size(), 0.0);

    for (int i = 0; i < fields.size(); ++i) {
        data[i] = fields.at(i).toDouble();
    }

    return data;
}

int BgmnDiaImport::countMultiScans(const QList<QByteArray> &lst) const
{
    if (lst.size() == 0) return 0;
    QByteArray first = lst.first().left(8);

    for (int i = 1; i < lst.size(); ++i) {
        if (lst.at(i).left(8) != first) return i;
    }

    return 0;
}

QList<QVector<double> > BgmnDiaImport::mergeMultiScans(const QList<QVector<double> > &data, int mult)
{
    QList<QVector<double> > out;
    out.reserve(data.size() / mult);

    int n = 0;
    while (n <= (data.size() - mult)) {
        QVector<double> line(data.at(n).size(), 0.0);
        line[0] = data.at(n).first();

        for (int i = 0; i < mult; ++i) {
            for (int v = 1; v < line.size(); ++v) {
                line[v] += data.at(n+i).at(v) / double(mult);
            }
        }

        out.append(line);
        n += mult;
    }

    return out;
}
