/***************************************************************************
                          savtemplatemanager.cpp  -  description
                             -------------------
    begin                : Mon Nov 22 19:02:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QRegularExpression>
#include <QFile>

#include "savtemplatemanager.h"
#include "../libXrdIO/bgmnfileio.h"

SavTemplateManager::SavTemplateManager()
{

}

void SavTemplateManager::loadTemplateFile(const QString &s)
{
    if (!QFile::exists(s)) {
        _content = defaultContent();
        _fileName = QFileInfo();
    } else {
        _content = BgmnFileIO::readTextFile(s);
        _fileName = QFileInfo(s);
    }
}

void SavTemplateManager::generateDefaultTemplate()
{
    _content = defaultContent();
    _fileName = QFileInfo();
}

void SavTemplateManager::saveTemplate(const QString &s)
{
    if (!s.isEmpty()) {
        _fileName = QFileInfo(s);
    }

    BgmnFileIO::writeTextFile(_fileName.absoluteFilePath(), _content);
}

double SavTemplateManager::getSynchrotron() const
{
    QRegularExpression rx("(?<!%)\\s*SYNCHROTRON=(\\d+\\.?\\d*)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return rm.captured(1).toDouble();
    }

    return -1.0;
}

QString SavTemplateManager::getLambda() const
{
    QRegularExpression rx("(?<!%)\\s*LAMBDA=(\\S+)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return rm.captured(1);
    }

    return QString();
}

QString SavTemplateManager::getBackground() const
{
    QRegularExpression rx("(?<!%)\\s*UNT=(\\S+)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return _fileName.absolutePath() + QDir::separator() + rm.captured(1);
    }

    return QString();
}

QString SavTemplateManager::getBackgroundC() const
{
    QRegularExpression rx("(?<!%)\\s*UNTC=(\\S+)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return _fileName.absolutePath() + QDir::separator() + rm.captured(1);
    }

    return QString();
}

QString SavTemplateManager::getTubeTails() const
{
    QRegularExpression rx("(?<!%)\\s*TubeTails=(\\S+)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return _fileName.absolutePath() + QDir::separator() + rm.captured(1);
    }

    return QString();
}

int SavTemplateManager::getRU() const
{
    QRegularExpression rx("(?<!%)\\s*RU=(\\S+)");
    QRegularExpressionMatch rm = rx.match(_content);

    if (rm.hasMatch()) {
        return rm.captured(1).toInt();
    }

    return -1;
}

void SavTemplateManager::setSynchrotron(double d)
{
    QRegularExpression rx("SYNCHROTRON=\\d*\\.?\\d*");
    _content.remove(QRegularExpression("LAMBDA=\\S*"));

    if (_content.contains(rx)) {
        _content.replace(rx, QString("SYNCHROTRON=%1").arg(d, 0, 'd', 8));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("% Wavelength\nSYNCHROTRON=%1\n").arg(d, 0, 'd', 8));
    }
}

void SavTemplateManager::setLambda(const QString &s)
{
    QRegularExpression rx("LAMBDA=\\S*");
    _content.remove(QRegularExpression("SYNCHROTRON=\\S*"));

    if (_content.contains(rx)) {
        _content.replace(rx, QString("LAMBDA=%1").arg(s));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("% Wavelength\nLAMBDA=%1\n").arg(s));
    }
}

void SavTemplateManager::setBackgroundC(const QString &s)
{
    QFileInfo fi(s);
    QRegularExpression rx("UNTC=\\S*");
    if (_content.contains(rx)) {
        _content.replace(rx, QString("UNTC=%1").arg(fi.fileName()));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("% Measured background\nUNTC=%1\n").arg(fi.fileName()));
    }
}

void SavTemplateManager::setBackground(const QString &s)
{
    QFileInfo fi(s);
    QRegularExpression rx("UNT=\\S*");
    if (_content.contains(rx)) {
        _content.replace(rx, QString("UNT=%1").arg(fi.fileName()));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("% Measured background\nUNT=%1\n").arg(fi.fileName()));
    }
}

void SavTemplateManager::setTubeTails(const QString &s)
{
    QFileInfo fi(s);
    QRegularExpression rx("TubeTails=\\S*");
    if (_content.contains(rx)) {
        _content.replace(rx, QString("TubeTails=%1").arg(fi.fileName()));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("% Measured tube tails\nTubeTails=%1\n").arg(fi.fileName()));
    }
}

void SavTemplateManager::setRU(int n)
{
    QRegularExpression rx("RU=\\d*");
    if (_content.contains(rx)) {
        _content.replace(rx, QString("RU=%1").arg(n));
    } else {
        int idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
        _content.insert(idx < 0 ? 0 : idx, QString("RU=%1\n").arg(n));
    }
}

void SavTemplateManager::setPolarization(bool b, const QString &ang)
{
    if (b) {
        int idx = _content.indexOf(QRegularExpression("POL="));

        if (idx >= 0) {
            _content.remove(QRegularExpression("POL=[^\\n]+\\n"));
            _content.remove(QRegularExpression("pi=[^\\n]+\\n"));
        } else {
            idx = _content.indexOf(QRegularExpression("%\\s+Phases|STRUC\\[\\d+\\]"));
            if (idx < 0) idx = 0;
        }

        _content.insert(idx, ang);
    } else {
        _content.remove(QRegularExpression("POL=[^\\n]+\\n"));
        _content.remove(QRegularExpression("pi=[^\\n]+\\n"));
    }
}

QString SavTemplateManager::defaultContent()
{
    QString s;

    s += QString("% Theoretical instrumental function\n");
    s += QString("VERZERR=\n");
    s += QString("% Wavelength\n");
    s += QString("LAMBDA=\n");
    s += QString("% Phases\n");
    s += QString("% Measured background\n");
    s += QString("% Measured data\n");
    s += QString("VAL[1]=\n");
    s += QString("% Minimum Angle (2theta)\n");
    s += QString("% WMIN=10\n");
    s += QString("% Maximum Angle (2theta)\n");
    s += QString("% WMAX=60\n");
    s += QString("% Result list output\n");
    s += QString("LIST=\n");
    s += QString("% Peak list output\n");
    s += QString("OUTPUT=\n");
    s += QString("% Diagram output\n");
    s += QString("DIAGRAMM=\n");
    s += QString("% Global parameters for zero point and sample displacement\n");
    s += QString("EPS1=0\n");
    s += QString("PARAM[1]=EPS2=0_-0.01^0.01\n");
    s += QString("EPS3=0\n");
    s += QString("alpha3ratio=0.020\n");
    s += QString("betaratio=0.005\n");
    s += QString("NTHREADS=8\n");
    s += QString("PROTOKOLL=Y\n");
    s += QString("SAVE=N\n");

    return s;
}
