/***************************************************************************
                          bgmnstrindexer.cpp  -  description
                             -------------------
    begin                : Sun Oct 31 13:14:00 CEST 2021
    copyright            : (C) 2021 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QDir>
#include "bgmnstrverifier.h"
#include "../libXrdIO/bgmnfileio.h"
#include "../libXrdIO/functions.h"
#include "../libXrdIO/parser/bgmnparparser.h"

BgmnStrVerifier::BgmnStrVerifier(QObject *parent)
    : QObject(parent)
{
    settings = SettingsManager::getInstance();
    bgmnVerifyHandler = nullptr;
}

BgmnStrVerifier::~BgmnStrVerifier()
{
    if (bgmnVerifyHandler) delete bgmnVerifyHandler;
}

void BgmnStrVerifier::runStrVerification(const QString &strString)
{
    QString oFile(QDir::fromNativeSeparators(settings->getTempLocation() + "/profexDummy"));
    indexedStructure.setStrFile(oFile + ".str");
    indexedStructure.setParFile(oFile + ".par");
    indexedStructure.setLstFile(oFile + ".lst");
    indexedStructure.setRemove(false);

    // remove old dummy files
    if (QFile::exists(indexedStructure.strFile().absoluteFilePath())) QFile::remove(indexedStructure.strFile().absoluteFilePath());
    if (QFile::exists(indexedStructure.parFile().absoluteFilePath())) QFile::remove(indexedStructure.parFile().absoluteFilePath());
    if (QFile::exists(indexedStructure.lstFile().absoluteFilePath())) QFile::remove(indexedStructure.lstFile().absoluteFilePath());

    if (!BgmnFileIO::writeTextFile(indexedStructure.strFile().absoluteFilePath(), strString)) {
        indexedStructure.setMessage(QString("Could not create temporary file in %1").arg(indexedStructure.strFile().absoluteFilePath()));
    } else {
        runVerification();
    }
}

void BgmnStrVerifier::runVerification()
{
    double wl = 0.1 * settings->defaultWavelength();
    double dmax = settings->value("config/hklUpperRangeD", 0.1540598).toDouble();
    double wmax = global::Functions::dToTwoTheta(dmax, wl);

    lastLogMessage = QString();

    if (!bgmnVerifyHandler) {
        bgmnVerifyHandler = new BgmnHandler(QUuid::createUuid());
        connect(bgmnVerifyHandler, SIGNAL(allComplete(QUuid)), this, SLOT(verifyProcessComplete()));
        //connect(bgmnVerifyHandler, SIGNAL(pollOutput()), this, SLOT(pollProcessOutput()));
    }

    if (bgmnVerifyHandler->init()) {
        QStringList args;

        args << indexedStructure.strFile().completeBaseName();
        args << "SAVE=N";
        args << QString("SYNCHROTRON=%1").arg(wl);
        args << QString("NTHREADS=%1").arg(QThread::idealThreadCount());
        args << QString("WMAX=%1").arg(wmax, 0, 'f', 2);
        args << QString("STRUC[1]=%1").arg(indexedStructure.strFile().absoluteFilePath());
        args << QString("OUTPUT=%1").arg(indexedStructure.parFile().absoluteFilePath());
        args << QString("LIST=%1").arg(indexedStructure.lstFile().absoluteFilePath());

        bgmnVerifyHandler->run(settings->getTempLocation(), args);
    } else {
        if (settings->verboseLevel() > 1) qDebug() << QString("BgmnRefStructureManager::runVerification(): Error, could not initialize bgmnHandler, exiting");
        verifyProcessComplete();
    }
}

void BgmnStrVerifier::verifyProcessComplete()
{
    if (!indexedStructure.parFile().exists()) {
        indexedStructure.reset();
        if (settings->verboseLevel() > 1) qDebug() << QString("BgmnStrVerifier::verifyProcessComplete(): Error: *.par file not created");
    } else {
        QVector<Hkl> vec = parseParFile(indexedStructure.parFile().absoluteFilePath());

        indexedStructure.hklPhase().setComment(getComment(indexedStructure.hklPhase().file()));
        indexedStructure.hklPhase().setPhase(vec.first().phase());
        indexedStructure.hklPhase().setColor(vec.first().color().name());
        indexedStructure.hklPhase().setHklData(vec);
        indexedStructure.hklPhase().setDensity(parseLstFile(indexedStructure.lstFile().absoluteFilePath()).value("XrayDensity", -1.0).toDouble());
    }

    if (indexedStructure.doRemove()) {
        if (indexedStructure.strFile().exists()) QFile::remove(indexedStructure.strFile().absoluteFilePath());
        if (indexedStructure.lstFile().exists()) QFile::remove(indexedStructure.lstFile().absoluteFilePath());
        if (indexedStructure.parFile().exists()) QFile::remove(indexedStructure.parFile().absoluteFilePath());
    }

    emit signalVerifyComplete();
}

QVector<Hkl> BgmnStrVerifier::parseParFile(const QString &parFile)
{
    bool b;
    BgmnParParser pparser(parFile, &b);

    if (!b) {
        if (settings->verboseLevel() > 1) qDebug() << QString("BgmnStrVerifier::parseParFile: BgmnParParser reported an error parsing file %1").arg(parFile);
        return QVector<Hkl>();
    }

    QVector<Hkl> vec = pparser.getReflections(); // only of first phase?
    if (settings->verboseLevel() > 1) qDebug() << QString("BgmnStrVerifier::parseParFile: *.par file was successfully parsed. Read %1 hkl indices").arg(vec.size());

    return vec;
}

QMap<QString, QVariant> BgmnStrVerifier::parseLstFile(const QString &lstFile)
{
    QMap<QString, QVariant> data;

    QString lstFileData(BgmnFileIO::readTextFile(lstFile));

    if (!lstFileData.isEmpty()) {
        QRegularExpression rxDensity("XrayDensity=(\\d+\\.?\\d*)");
        QRegularExpressionMatch rmDensity = rxDensity.match(lstFileData);
        if (rmDensity.hasMatch()) {
            data["XrayDensity"] = rmDensity.captured(1).toDouble();
        } else {
            data["XrayDensity"] = -1.0;
        }
    }

    return data;
}

QString BgmnStrVerifier::getComment(const QString &file)
{
    QRegularExpression rx("^PHASE=([^/]+)//(.*)$");

    QFile f(file);

    if (f.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QTextStream in(&f);

        while (!in.atEnd()) {
            QRegularExpressionMatch rm = rx.match(in.readLine());

            if (rm.hasMatch()) {
                return rm.captured(2).trimmed();
            }
        }
    }

    f.close();
    return QString("");
}
