/***************************************************************************
                          chemtabledata.cpp  -  description
                             -------------------
    begin                : Sun Feb 21 13:41:07 CEST 2016
    copyright            : (C) 2016 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "chemtabledata.h"
#include <QDebug>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <QtMath>

ChemTableData::ChemTableData()
{
    mode = OXIDE;
    settings = SettingsManager::getInstance();
}

void ChemTableData::readSettings()
{
    mode = static_cast<ChemistryMode>(settings->value("chemistry/mode", 0).toInt());
    QStringList oxideSettingsData = settings->value("chemistry/oxides", QStringList()).toStringList();

    if (oxideSettingsData.size() < 3) {
        qDebug() << QStringLiteral("ChemTableData::readSettings(): StringList is too short.");
        oxideSettingsData = defaultOxideWeights();
    }

    initElementList(defaultElementWeights());
    initOxideList(oxideSettingsData);
}

void ChemTableData::initElementList(const QStringList &elementData)
{
    if (elementData.isEmpty()) return;

    atomElementLookupTable.clear();
    elementNumberLookupTable.clear();

    for (int i = 0; i < elementData.size() - 1; i += 2) {
        QString atomName     = elementData.at(i);
        double elementWeight = elementData.at(i+1).toDouble();
        int elementNumber    = elementNumberLookupTable.size();

        elementNumberLookupTable.insert(atomName, elementNumber);

        if (elementWeight > 0.0) {
            atomElementLookupTable.insert(atomName.toUpper(), Oxide(atomName, elementWeight, 0.0));
        } else {
            qDebug() << QString("ChemTableData::initElementList(): No molecular weight defined for element %1").arg(atomName);
        }
    }
}

void ChemTableData::initOxideList(const QStringList &oxideSettingsData)
{
    atomOxideLookupTable.clear();
    oxideNumberLookupTable.clear();
    QStringList missing;

    for (int i = 0; i < oxideSettingsData.size() - 2; i += 3) {
        QString missingEl = appendOxideToLUTs(oxideSettingsData.mid(i, 3));
        if (!missingEl.isEmpty()) missing << missingEl;
    }

    if (missing.size()) {
        qDebug() << QString("ChemTableData::initOxideList(): No oxide weight defined in preferences for elements %1").arg(missing.join(", "));
    }
}

QString ChemTableData::appendOxideToLUTs(const QStringList &oxideData)
{
    if (oxideData.isEmpty()) return QString();

    QString atomName    = oxideData.at(0);
    QString oxideName   = oxideData.size() > 1 ? oxideData.at(1)            : atomName;
    double  oxideWeight = oxideData.size() > 2 ? oxideData.at(2).toDouble() : 0.0;
    int     oxideNumber = oxideNumberLookupTable.size();

    if (atomName.trimmed() == "O") return QString();

    oxideNumberLookupTable.insert(oxideName, oxideNumber);

    if (oxideWeight > 0.0) {
        atomOxideLookupTable.insert(atomName.toUpper(), Oxide(oxideName, oxideWeight, 0.0));
        return QString();
    }

    return atomName;
}

void ChemTableData::setAtomsAndGoals(const QMultiMap<Phase, CrystalStructure> &phasesStructures)
{
    phaseWeightData = structuresToAtoms(phasesStructures);
}

QMultiMap<Phase, Oxide> ChemTableData::structuresToAtoms(const QMultiMap<Phase, CrystalStructure> &structures)
{
    QMultiMap<Phase, Oxide> phaseWeightList;
    QStringList phases = structures.keys();

    for (int i = 0; i < phases.size(); ++i) {
        CrystalStructure phaseStructure = structures.value(phases.at(i));
        QList<Oxide> phaseDataNormalized;

        switch (mode) {
        case ELEMENT:
            phaseDataNormalized = atomsToElementsNormalized(phaseStructure.atoms());
            break;
        case ATOMIC:
            phaseDataNormalized = atomsToAtomicNormalized(phaseStructure.atoms());
            break;
        case OXIDE:
            phaseDataNormalized = atomsToOxidesNormalized(phaseStructure.atoms());
            break;
        default:
            phaseDataNormalized = atomsToElementsNormalized(phaseStructure.atoms());
            break;
        }

        // store the results in oxResults
        for (int j = 0; j < phaseDataNormalized.size(); ++j) {
            // insert the results, multiple keys required
            phaseWeightList.insert(phases.at(i), phaseDataNormalized.at(j));
        }
    }

    return phaseWeightList;
}

QList<Oxide> ChemTableData::atomsToOxidesNormalized(const QList<CrystalAtom> &atomList)
{
    double sum = 0.0;
    QHash<QString, Oxide> oxideList;

    for (int i = 0; i < atomList.size(); ++i) {
        CrystalAtom atom = atomList.at(i);

        if (!atomOxideLookupTable.contains(atom.element())) {
            continue;
        }

        QString oxideName = atomOxideLookupTable[atom.element().toUpper()].name;

        // extract the number of cations, e.g. the "2" in "Al2O3",
        // because we must divide the oxide weight by this number

        double sc = 1.0;
        QRegularExpression rx("([A-Z][a-z]?)(\\d*)O?\\d*");
        QRegularExpressionMatch match = rx.match(oxideName);

        if (match.hasMatch()) {
            bool ok;
            sc = match.captured(2).toDouble(&ok);
            if (!ok) sc = 1.0;
        }

        double val = atom.occupancy() * atom.multiplicity() * atomOxideLookupTable[atom.element()].weight / (qFuzzyIsNull(sc) ? 1.0 : sc);
        double esd = atom.occupancy_esd() * atom.multiplicity() * atomOxideLookupTable[atom.element()].weight / (qFuzzyIsNull(sc) ? 1.0 : sc);
        sum += val;

        if (oxideList.contains(oxideName)) {
            oxideList[oxideName].weight += val;
            oxideList[oxideName].esd = qSqrt(qPow(oxideList[oxideName].esd, 2.0) + qPow(esd, 2.0));
        } else {
            oxideList.insert(oxideName, Oxide(oxideName, val, esd));
        }
    }

    // normalize to sum
    QHash<QString, Oxide>::iterator it = oxideList.begin();
    while (it != oxideList.end()) {
        it.value().weight /= sum;
        it.value().esd    /= sum;
        ++it;
    }

    return oxideList.values();
}

QList<Oxide> ChemTableData::atomsToElementsNormalized(const QList<CrystalAtom> &atomList)
{
    double sum = 0.0;
    QHash<QString, Oxide> elementList;

    for (int i = 0; i < atomList.size(); ++i) {
        CrystalAtom atom = atomList.at(i);

        if (!atomElementLookupTable.contains(atom.element())) {
            continue;
        }

        QString elementName = atomElementLookupTable[atom.element().toUpper()].name;

        double val = atom.occupancy() * atom.multiplicity() * atomElementLookupTable[atom.element()].weight;
        double esd = atom.occupancy_esd() * atom.multiplicity() * atomElementLookupTable[atom.element()].weight;
        sum += val;

        if (elementList.contains(elementName)) {
            elementList[elementName].weight += val;
            elementList[elementName].esd = qSqrt(qPow(elementList[elementName].esd, 2.0) + qPow(esd, 2.0));
        } else {
            elementList.insert(elementName, Oxide(elementName, val, esd));
        }
    }

    // normalize to sum
    QHash<QString, Oxide>::iterator it = elementList.begin();
    while (it != elementList.end()) {
        it.value().weight /= sum;
        it.value().esd    /= sum;
        ++it;
    }

    return elementList.values();
}

QList<Oxide> ChemTableData::atomsToAtomicNormalized(const QList<CrystalAtom> &atomList)
{
    double sum = 0.0;
    QHash<QString, Oxide> elementList;

    for (int i = 0; i < atomList.size(); ++i) {
        CrystalAtom atom = atomList.at(i);

        QString elementName = atomElementLookupTable[atom.element().toUpper()].name;

        double val = atom.occupancy() * atom.multiplicity();
        double esd = atom.occupancy_esd() * atom.multiplicity();
        sum += val;

        if (elementList.contains(elementName)) {
            elementList[elementName].weight += val;
            elementList[elementName].esd = qSqrt(qPow(elementList[elementName].esd, 2.0) + qPow(esd, 2.0));
        } else {
            elementList.insert(elementName, Oxide(elementName, val, esd));
        }
    }

    // normalize to sum
    QHash<QString, Oxide>::iterator it = elementList.begin();
    while (it != elementList.end()) {
        it.value().weight /= sum;
        it.value().esd    /= sum;
        ++it;
    }

    return elementList.values();
}

void ChemTableData::clear()
{
    phaseWeightData.clear();
}

QStringList ChemTableData::defaultOxideWeights()
{
    QStringList atomList = global::atoms.split(";", Qt::KeepEmptyParts);
    QStringList defaultOxideList;

    for (int i = 0; i < atomList.size() - 4; i += 5) {
        defaultOxideList << atomList.at(i+1) << atomList.at(i+3) << atomList.at(i+4);
    }

    return defaultOxideList;
}

QStringList ChemTableData::defaultElementWeights()
{
    QStringList atomList = global::atoms.split(";", Qt::KeepEmptyParts);
    QStringList defaultElementList;

    for (int i = 0; i < atomList.size() - 4; i += 5) {
        defaultElementList << atomList.at(i+1) << atomList.at(i+2);
    }

    return defaultElementList;
}

QStringList ChemTableData::getAllPhaseNames()
{
    QStringList phNames = phaseWeightData.keys();
    phNames.removeDuplicates();
    return phNames;
}

QStringList ChemTableData::getAllOxideNames()
{
    QMap<int, QString> namesSorted;
    QList<Oxide> elements = phaseWeightData.values();

    for (int i = 0; i < elements.size(); ++i) {
        int n = 0;

        switch (mode) {
        case ELEMENT:
            n = elementNumberLookupTable.value(elements.at(i).name);
            break;
        case OXIDE:
            n = oxideNumberLookupTable.value(elements.at(i).name);
            break;
        case ATOMIC:
            n = elementNumberLookupTable.value(elements.at(i).name);
            break;
        default:
            n = elementNumberLookupTable.value(elements.at(i).name);
            break;
        }

        namesSorted.insert(n, elements.at(i).name);
    }

    return namesSorted.values();
}

QList<Oxide> ChemTableData::getPhaseChemistry(const QString &phase)
{
    return phaseWeightData.values(phase);
}

QHash<QString, double> ChemTableData::getPhaseChemWeights(const QString &phase)
{
    QList<Oxide> phOxides = phaseWeightData.values(phase);
    QHash<QString, double> oxHash;

    for (int i = 0; i < phOxides.size(); ++i) {
        oxHash.insert(phOxides.at(i).name, phOxides.at(i).weight);
    }

    return oxHash;
}


QHash<QString, double> ChemTableData::getPhaseChemEsds(const QString &phase)
{
    QList<Oxide> phOxides = phaseWeightData.values(phase);
    QHash<QString, double> oxHash;

    for (int i = 0; i < phOxides.size(); ++i) {
        oxHash.insert(phOxides.at(i).name, phOxides.at(i).esd);
    }

    return oxHash;
}
