/***************************************************************************
                          chemtablewidget.cpp  -  description
                             -------------------
    begin                : Mon Jun 30 14:16:07 CEST 2014
    copyright            : (C) 2014 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ui_chemtablewidgetform.h"
#include "chemtablewidget.h"
#include "../libXrdIO/crystal/crystalstructure.h"
#include <QDebug>
#include <QComboBox>
#include <QHeaderView>
#include <QtMath>
#include <QMimeData>
#include <QApplication>
#include <QClipboard>
#include <QFileInfo>
#include <QLocale>

ChemTableWidget::ChemTableWidget(GraphDataController *c, QWidget *parent) :
    AbstractGraphView(c, parent),
    ui(new Ui::ChemTableWidgetForm)
{
    ui->setupUi(this);
    ui->tableWidget->setEditTriggers(QAbstractItemView::NoEditTriggers);
    ui->tableWidget->setSelectionMode(QAbstractItemView::ContiguousSelection);
    ui->tableWidget->setSelectionBehavior(QAbstractItemView::SelectItems);

    contextMenu = new QMenu(this);
    contextMenu->addAction(tr("Copy"), this, SLOT(copy()));
    connect(ui->tableWidget->horizontalHeader(), SIGNAL(sectionResized(int,int,int)), this, SLOT(headerResized()));
}

ChemTableWidget::~ChemTableWidget()
{
    if (contextMenu) delete contextMenu;
    if (ui) delete ui;
}

void ChemTableWidget::initSettings()
{
    oxideData.readSettings();
    mode = static_cast<ChemistryMode>(settings->value("chemistry/mode", 0).toInt());
    if (fileName.isEmpty()) setData(fileName);
}

bool ChemTableWidget::hasData()
{
    return ((ui->tableWidget->rowCount() > 0) && (ui->tableWidget->columnCount() > 0)) ? true : false;
}

void ChemTableWidget::updateView()
{
    /* We do not react to view updates. Call ::setData() manually
     * if new data is available */
}

void ChemTableWidget::setShown(bool b)
{
    isShown = b;
    if (!fileName.isEmpty()) setData(fileName);
}

void ChemTableWidget::setData(const QString &lstFile, bool force)
{
    if (!isShown && !force) return;

    bool ok;
    BgmnLstParser lparser(lstFile, ok);
    if (ok) setData(lparser, force);
}

void ChemTableWidget::setData(const BgmnLstParser &lparser, bool force)
{
    QFileInfo fi(lparser.getFileName());
    fileName = fi.absoluteFilePath();
    sampleName = fi.completeBaseName();
    if (!isShown && !force) return;

    previousManualGoalAssignments = settings->value("bgmnProject/chemGoalAssignments", QVariant()).toHash();

    QStringList phaseNames = lparser.getPhaseNames();
    QMultiMap<QString, CrystalStructure> phaseStructures;

    for (int i = 0; i < phaseNames.size(); ++i) {
        phaseStructures.insert(phaseNames.at(i), lparser.getCrystalStructure(phaseNames.at(i)));
    }

    oxideData.setAtomsAndGoals(phaseStructures);
    resizeTable();

    QStringList globalGoals = settings->value("bgmnProject/reportedGlobalGoals", global::defaultBgmnGlobalGoals.split("\n")).toStringList();
    if (globalGoals.isEmpty()) globalGoals << ".*";

    createPhaseComboBoxes(lparser.getGlobalGoals(globalGoals));
    recalculate();
}

bool ChemTableWidget::reload()
{
    if (fileName.isEmpty()) return false;

    setData(fileName, true);
    return true;
}

void ChemTableWidget::resizeTable()
{
    QByteArray hState = ui->tableWidget->columnCount() > 0
              ? ui->tableWidget->horizontalHeader()->saveState()
              : settings->value("config/chemTableHeader", ui->tableWidget->horizontalHeader()->saveState()).toByteArray();

    clearAll();

    QStringList phaseNames = oxideData.getAllPhaseNames();
    if (!phaseNames.size()) return;

    QStringList oxideNames = oxideData.getAllOxideNames();

    QStringList hHeaderLabels;
    hHeaderLabels << tr("Quantity Goal") << tr("Phase Quantity (wt-%)");

    QString unit = mode == ATOMIC ? " (atm-%)" : " (wt-%)";

    for (int i = 0; i < oxideNames.size(); ++i) {
        hHeaderLabels << oxideNames.at(i) + unit;
    }

    QStringList vHeaderLabels;
    vHeaderLabels.append(phaseNames);
    vHeaderLabels.append("Weighted total");

    ui->tableWidget->setColumnCount(hHeaderLabels.size());
    ui->tableWidget->setRowCount(vHeaderLabels.size());

    ui->tableWidget->setHorizontalHeaderLabels(hHeaderLabels);
    ui->tableWidget->setVerticalHeaderLabels(vHeaderLabels);

    bool oldState = ui->tableWidget->blockSignals(true);
    ui->tableWidget->horizontalHeader()->restoreState(hState);
    ui->tableWidget->blockSignals(oldState);
}

void ChemTableWidget::createPhaseComboBoxes(const QList<Result> &goals)
{
    bool div = settings->value("bgmnProject/quantGoals100percent", false).toBool();

    QMap<QString, double> goalValues;

    for (int i = 0; i < goals.size(); ++i) {
        Result result = goals.at(i);
        goalValues.insert(result.name, div?  result.value / 100.0 : result.value);
    }

    for (int i = 0; i < ui->tableWidget->rowCount() - 1; ++i) {
        QComboBox *cbox = new QComboBox(this);

        ui->tableWidget->setCellWidget(i, 0, cbox);
        cbox->addItem("-", QVariant(0.0));

        QMap<QString, double>::ConstIterator it = goalValues.constBegin();

        while (it != goalValues.constEnd()) {
            cbox->addItem(it.key(), QVariant(it.value()));
            ++it;
        }

        selectComboBoxItem(cbox, ui->tableWidget->verticalHeaderItem(i)->text());
        connect(cbox, SIGNAL(currentIndexChanged(int)), this, SLOT(recalculate()));
    }
}

bool ChemTableWidget::selectComboBoxItem(QComboBox *comboBox, const QString &string)
{
    if (previousManualGoalAssignments.contains(string)) {
        comboBox->setCurrentText(previousManualGoalAssignments.value(string).toString());
        return true;
    }

    QRegularExpression rx(QString("(?:100\\*|Q)?(?:%1)(?:\\/sum)?").arg(string));

    for (int i = 0; i < comboBox->count(); ++i) {
        QRegularExpressionMatch rm = rx.match(comboBox->itemText(i));

        if (rm.hasMatch()) {
            comboBox->setCurrentIndex(i);
            return true;
        }
    }

    return false;
}

void ChemTableWidget::recalculate()
{
    updateManualGoalAssignments();
    calculateQuantityCells();
    calculatePhaseCells();

    if (mode != ATOMIC) {
        calculateTotalCells();
    } else {
        clearTotalCells();
    }
}

QMap<int, double> ChemTableWidget::normalizeGoalValues()
{
    double sum = 0.0;

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(r, 0));
        if (!cbox) continue;
        sum += cbox->currentData(Qt::UserRole).toDouble();
    }

    double factor = 100.0;
    if (qAbs(sum - 0.01)  < 0.0001) factor = 10000.0;
    if (qAbs(sum - 100.0) < 1.00)   factor = 1.0;

    QMap<int, double> values;

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(r, 0));
        if (!cbox) continue;
        values[r] = factor * cbox->currentData(Qt::UserRole).toDouble();
    }

    return values;
}

/*
 * old implementation
void ChemTableWidget::calculateQuantityCells()
{
    bool allAssigned = true;
    QMap<int, double> cellQuantValues = normalizeGoalValues();

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(r, 0));
        if (!cbox) continue;

        if (cbox->currentIndex() == 0) {
            allAssigned = false;
            setCellText(r, 1, "-");
        } else {
            setCellValue(r, 1, cellQuantValues.value(r), 0.0, QColor(), 2);
        }
    }

    if (allAssigned) {
        double sum = 0.0;

        for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
            sum += getCellValue(r, 1, 0.0);
        }

        double loLimitPercent = 100.0 - 0.05 * ui->tableWidget->rowCount();
        double upLimitPercent = 100.0 + 0.05 * ui->tableWidget->rowCount();

        bool sumOk = ((sum < upLimitPercent) && (sum > loLimitPercent));
        QColor col(sumOk ? "#d1e9f8" : "#ffcccc");

        setCellValue(ui->tableWidget->rowCount() - 1, 1, sum, sumOk ? 1.0 : -1.0, col, 2);
    } else {
        setCellText(ui->tableWidget->rowCount() - 1, 1, "-", QColor("#d1e9f8"));
    }
}
*/

/*
 * This is a new implementation. It calculates the quantities also when the sum is different from 100%.
 * The reason is that the checks for correct goal assignement (checks for sum=100%) did not work
 * with absolute phase quantities / internal standard.
 */
void ChemTableWidget::calculateQuantityCells()
{
    QMap<int, double> cellQuantValues = normalizeGoalValues();

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(r, 0));
        if (!cbox) continue;

        if (cbox->currentIndex() == 0) {
            setCellText(r, 1, "-");
        } else {
            setCellValue(r, 1, cellQuantValues.value(r), 0.0, QColor(), 2);
        }
    }

    double sum = 0.0;

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        sum += getCellValue(r, 1, 0.0);
    }

    setCellValue(ui->tableWidget->rowCount() - 1, 1, sum, 1.0, QColor(), 2);
}

void ChemTableWidget::calculatePhaseCells()
{
    for (int i = 0; i < ui->tableWidget->rowCount() - 1; ++i) {
        QString phase = ui->tableWidget->verticalHeaderItem(i)->text();
        QHash<QString, double> oxideWeights = oxideData.getPhaseChemWeights(phase);
        QHash<QString, double> oxideEsds    = oxideData.getPhaseChemEsds(phase);

        // now loop over all oxides found in the horizontal header, and
        // check if it is found in the hash. if yes, insert the value
        // if no, insert 0 into the table
        static QRegularExpression rxOxide("([A-Za-z0-9]+)");
        QRegularExpressionMatch rmOxide;

        for (int j = 2; j < ui->tableWidget->columnCount(); ++j) {
            double value = 0.0;
            double esd   = 0.0;

            rmOxide = rxOxide.match(ui->tableWidget->horizontalHeaderItem(j)->text());

            if (rmOxide.hasMatch()){
                QString oxName = rmOxide.captured(1);
                value = oxideWeights.value(oxName, 0.0) * 100.0;
                esd   = oxideEsds.value(oxName, 0.0) * 100.0;
            }

            setCellValue(i, j, value, esd);
        }
    }
}

double ChemTableWidget::getPhaseQuantity(int row)
{
    if (!cellIsValid(row, 0)) return 0.0;

    QWidget *cWidget = ui->tableWidget->cellWidget(row, 0);
    if (!cWidget) return 0.0;

    QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(row, 0));
    QVariant cdata = cbox->currentData();

    if (!cdata.isValid()) return 0.0;
    return cdata.toDouble();
}

void ChemTableWidget::setCellValue(int row, int column, double value, double esd, const QColor &color, int digits)
{
    if (!cellIsValid(row, column)) return;

    QTableWidgetItem *it = ui->tableWidget->item(row, column);
    QString valueString = value < 0.0 ? QString("-") : QString("%L1").arg(value, 0, 'f', digits);

    if (!it) {
        it = new QTableWidgetItem();
        ui->tableWidget->setItem(row, column, it);
    }

    it->setTextAlignment(Qt::AlignCenter);
    it->setText(valueString);
    it->setData(Qt::UserRole, QVariant(esd));

    if (color.isValid()) {
        it->setBackground(color);
    }
}

void ChemTableWidget::setCellText(int row, int column, const QString &txt, const QColor &color)
{
    if (!cellIsValid(row, column)) return;

    QTableWidgetItem *it = ui->tableWidget->item(row, column);

    if (!it) {
        it = new QTableWidgetItem();
        ui->tableWidget->setItem(row, column, it);
    }

    it->setTextAlignment(Qt::AlignCenter);
    it->setText(txt);

    if (color.isValid()) {
        it->setBackground(color);
    }
}

/*
 * old implmenentation
void ChemTableWidget::calculateTotalCells()
{
    bool valid = allQuantityGoalsSelected();

    for (int j = 2; j < ui->tableWidget->columnCount(); ++j) {
        if (valid) {
            double value = 0.0;
            double esd = 0.0;

            for (int i = 0; i < ui->tableWidget->rowCount() - 1; ++i) {
                double quantity = getPhaseQuantity(i);
                value += getCellValue(i, j, 0.0) * quantity;
                esd = qSqrt(qPow(esd, 2.0) + qPow(getCellEsd(i, j, 0.0), 2.0));
            }

            setCellValue(ui->tableWidget->rowCount() - 1, j, value, esd, QColor("#d1e9f8"));
        } else {
            setCellText(ui->tableWidget->rowCount() - 1, j, "-");
        }
    }
}
*/

/*
 * new implementation, does not check for correct quantity goals (sum = 100%), because
 * that didn't work with absolute phase quantities.
 */
void ChemTableWidget::calculateTotalCells()
{
    for (int j = 2; j < ui->tableWidget->columnCount(); ++j) {
        double value = 0.0;
        double esd = 0.0;

        for (int i = 0; i < ui->tableWidget->rowCount() - 1; ++i) {
            double quantity = getPhaseQuantity(i);
            value += getCellValue(i, j, 0.0) * quantity;
            esd = qSqrt(qPow(esd, 2.0) + qPow(getCellEsd(i, j, 0.0), 2.0));
        }

        setCellValue(ui->tableWidget->rowCount() - 1, j, value, esd, QColor("#d1e9f8"));
    }
}

void ChemTableWidget::clearTotalCells()
{
    for (int i = 1; i < ui->tableWidget->columnCount(); ++i) {
        setCellValue(ui->tableWidget->rowCount() - 1, i, -1.0, 0.0, QColor("#d1e9f8"));
    }
}

bool ChemTableWidget::cellIsValid(int row, int column)
{
    if (row    < 0)              return false;
    if (column < 0)              return false;
    if (row    >= ui->tableWidget->rowCount())    return false;
    if (column >= ui->tableWidget->columnCount()) return false;
    return true;
}

double ChemTableWidget::getCellValue(int row, int column, double defaultValue)
{
    if (!cellIsValid(row, column)) return defaultValue;

    QTableWidgetItem *it = ui->tableWidget->item(row, column);
    if (!it) return defaultValue;

    bool ok = false;
    QLocale locale(QLocale::system());
    double value = locale.toDouble(it->text(), &ok);
    return ok ? value : defaultValue;
}

double ChemTableWidget::getCellEsd(int row, int column, double defaultValue)
{
    if (!cellIsValid(row, column)) return defaultValue;

    QTableWidgetItem *it = ui->tableWidget->item(row, column);
    if (!it) return defaultValue;

    bool ok = false;
    double value = it->data(Qt::UserRole).toDouble(&ok);
    return ok ? value : defaultValue;
}

bool ChemTableWidget::allQuantityGoalsSelected()
{
    if (ui->tableWidget->rowCount() < 2) return false;

    // we check if the cell holding the sum of all quantities contains a
    // data value > 0.0. If it is negative or the cell item is null, the sum
    // is not 100%. If the item is not null and the data value is > 0.0, the
    // sum is 100% (or close enough).
    bool valueOk = true;

    QTableWidgetItem *it = ui->tableWidget->item(ui->tableWidget->rowCount() - 1, 1);

    if (!it) valueOk = false;
    if (it->data(Qt::UserRole).toDouble() < 0.0) valueOk = false;

    // we also check (as a second criterium) if all Quant Goal comboBoxes
    // are set to an index > 0, because phases refined to 0% would lead to
    // a correct sum even if the quant goal was not selected.
    bool assignOk = true;

    for (int r = 0; r < ui->tableWidget->rowCount() - 1; ++r) {
        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(r, 0));
        if (cbox) {
            if (cbox->currentIndex() == 0) assignOk = false;
        } else {
            assignOk = false;
        }
    }

    return (valueOk && assignOk);
}

/*
 * Returns the table in csv ";" list format. One parameter per line.
 * The header can be switched on or off, so the output of several
 * projects can be merged without header lines interrupting the lines
 */
QStringList ChemTableWidget::getCsvList(bool header, const QString &sampleId)
{
    if (!hasData()) {
        if (!reload()) return QStringList();
    }

    QString sep(";");
    QStringList out;

    if (header) {
        QStringList columnHeaders(QString("File"));
        columnHeaders << QString("Sample");
        columnHeaders << QString("Sample ID");
        columnHeaders << QString("Phase");
        columnHeaders << QString("Phase Quantity");

        switch (mode) {
        case ELEMENT:
            columnHeaders << QString("Element");
            columnHeaders << QString("Quantity (wt-%)");
            break;
        case ATOMIC:
            columnHeaders << QString("Element");
            columnHeaders << QString("Quantity (atm-%)");
            break;
        case OXIDE:
            columnHeaders << QString("Oxide");
            columnHeaders << QString("Quantity (wt-%)");
            break;
        default:
            columnHeaders << QString("Element");
            columnHeaders << QString("Quantity (wt-%)");
            break;
        }

        out += columnHeaders.join(sep);
    }

    QRegularExpression rx("([A-Za-z0-9\\s]+)(?:\\s+\\([a-z]+-%\\))");
    QRegularExpressionMatch rm;
    double sum = 0.0;

    // gather data
    for (int i = 0; i < ui->tableWidget->rowCount(); ++i) {
        double q = (i == ui->tableWidget->rowCount() - 1) ? -1.0 : getPhaseQuantity(i);
        sum += (i == ui->tableWidget->rowCount() - 1) ? 0.0 : q;

        for (int j = 2; j < ui->tableWidget->columnCount(); ++j) {
            rm = rx.match(ui->tableWidget->horizontalHeaderItem(j)->text());

            QString element;
            if (rm.hasMatch()) {
                element = rm.captured(1);
            } else {
                element = ui->tableWidget->horizontalHeaderItem(j)->text();
            }

            QStringList line(fileName);
            line << sampleName << sampleId;
            line << ui->tableWidget->verticalHeaderItem(i)->text();
            line << QString("%L1").arg(q < 0.0 ? sum : q, 0, 'f', 4);
            line << element;
            line << ui->tableWidget->item(i, j)->text();

            out.append(line.join(sep));
        }
    }

    return out;
}

void ChemTableWidget::clearAll()
{
    ui->tableWidget->clear();
    ui->tableWidget->setColumnCount(0);
    ui->tableWidget->setRowCount(0);
}

/*
 * Returns the table as csv (;) table sorted as shown on screen.
 */
QString ChemTableWidget::getCsvTable()
{
    if (!hasData()) {
        if (!reload()) return QString();
    }

    QString sep(";");
    QString out(fileName + "\n");
    QStringList line("Phase");

    for (int i = 0; i < ui->tableWidget->horizontalHeader()->count(); ++i) {
        line << ui->tableWidget->horizontalHeaderItem(i)->text();
    }

    out += line.join(sep) + "\n";

    for (int i = 0; i < ui->tableWidget->rowCount(); ++i) {
        line.clear();
        line << ui->tableWidget->verticalHeaderItem(i)->text();

        for (int j = 0; j < ui->tableWidget->columnCount(); ++j) {
            if (j == 0) {
                QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(i, j));
                if (cbox) line << cbox->currentText();
                else      line << "";
            } else {
                line << ui->tableWidget->item(i, j)->text();
            }
        }

        out += line.join(sep) + "\n";
    }

    return out;
}

/*
 * Returns the table in Html table format as shown on screen.
 */
QString ChemTableWidget::getHtmlTable()
{
    if (!hasData()) {
        if (!reload()) return QString();
    }

    QString output("<table id=\"tableChemComposition\">\n");

    output += QString("  <tr>\n");
    output += QString("    <th>Phase</th>\n");

    // skipping column 0, using the vertical header item instead
    for (int c = 1; c < ui->tableWidget->columnCount(); ++c) {
        output += QString("    <th>%1</th>\n").arg(ui->tableWidget->horizontalHeaderItem(c)->text());
    }

    output += QString("  </tr>\n");

    for (int r = 0; r < ui->tableWidget->rowCount(); ++r) {
        output += "  <tr>\n";

        for (int c = 0; c < ui->tableWidget->columnCount(); ++c) {
            if (c == 0) output += QString("    <td>%1</td>\n").arg(ui->tableWidget->verticalHeaderItem(r)->text());
            else        output += QString("    <td>%1</td>\n").arg(ui->tableWidget->item(r, c)->text());
        }

        output += "  </tr>\n";
    }

    output += "</table><br>\n";
    return output;
}

void ChemTableWidget::updateManualGoalAssignments()
{
    for (int i = 0; i < ui->tableWidget->rowCount(); ++i) {
        QWidget *wdg = ui->tableWidget->cellWidget(i, 0);
        if (!wdg) continue;

        QComboBox *cbox = static_cast<QComboBox*>(ui->tableWidget->cellWidget(i, 0));
        QString goal = cbox->currentText();

        if (goal != "-") {
            previousManualGoalAssignments[ui->tableWidget->verticalHeaderItem(i)->text()] = goal;
        }
    }

    settings->setValue("bgmnProject/chemGoalAssignments", QVariant(previousManualGoalAssignments));
}

void ChemTableWidget::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == Qt::RightButton) {
        contextMenu->exec(e->globalPosition().toPoint());
    }
}

void ChemTableWidget::copy()
{
    QMimeData *mime = new QMimeData();
    mime->setText(getCsvTable());
    qApp->clipboard()->setMimeData(mime);
}

/*
 * returns the phase quantity stored in a combobox'es data
 */
QString ChemTableWidget::getQuantityGoal(int i) const
{
    QWidget *cWidget = ui->tableWidget->cellWidget(i, 0);

    if (!cWidget) {
        return QString(" ");
    }

    QComboBox *cbox = static_cast<QComboBox*>(cWidget);

    if (cbox->currentIndex() == 0) {
        return QString(" ");
    }

    return cbox->currentText();
}

/*
 * returns the phase quantity stored in a combobox'es data
 */
QString ChemTableWidget::getQuantityValue(int i) const
{
    QWidget *cWidget = ui->tableWidget->cellWidget(i, 0);

    if (!cWidget) {
        return QString(" ");
    }

    QComboBox *cbox = static_cast<QComboBox*>(cWidget);

    if (cbox->currentIndex() == 0) {
        return QString(" ");
    }

    return cbox->currentData(Qt::UserRole).toString();
}

void ChemTableWidget::headerResized()
{
    settings->setValue("config/chemTableHeader", ui->tableWidget->horizontalHeader()->saveState());
}
