/***************************************************************************
                          controlfileedit.cpp  -  description
                             -------------------
    begin                : Thu Jul 08 08:00:00 CEST 2014
    copyright            : (C) 2014 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "controlfileedit.h"
#include "bgmnbackendconfig.h"

#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <QTextBlock>
#include <QDebug>
#include <QThread>
#include "../libXrdIO/structs.h"

ControlFileEdit::ControlFileEdit(QWidget *parent, const ControlFileEditFileType ft) :
    CodeEditor(parent), fileType(ft)
{
    settings = SettingsManager::getInstance();
    hasChangedText = false;
    maxNThreads = QThread::idealThreadCount();

    BgmnBackendConfig bgConfig;
    lamFiles = bgConfig.getAllLamFiles();

    // actions for the context menu
    actionSetRefinableIso = new QAction("Refine isotropically", this);
    actionSetRefinableAniso = new QAction("Refine anisotropically", this);
    actionSetRefinableFix = new QAction("Fix parameter", this);
    actionSetRefinableNumbered = new QAction("Refine parameter", this);
    actionOpenTextFile = new QAction("Open file as text", this);
    actionOpenGraphFile = new QAction("Open file as graph", this);

    actionAddStrucOutFile = new QAction("Add STRUCOUT file", this);
    actionAddSimpleStrucOutFile = new QAction("Add SimpleSTRUCOUT file", this);
    actionAddResFcfOutFile = new QAction("Add RESOUT and FCFOUT file", this);
    actionAddPdbOutFile = new QAction("Add PDBOUT file", this);

    // action for direct toggling via shortcut
    actionToggleRefineStateDown = new QAction("Decrease value", this);
    actionToggleRefineStateUp = new QAction("Increase value", this);

    actionCommentLine = new QAction("Comment line", this);
    actionUncommentLine = new QAction("Uncomment line", this);
    actionToggleCommentLine = new QAction("Toggle comment line", this);

    actionSetParameterNo = new QAction("Set to disabled", this);
    actionSetParameterYes = new QAction("Set to enabled", this);

    actionToggleRefineStateDown->setShortcut(Qt::Key_F5);
    actionToggleRefineStateUp->setShortcut(Qt::Key_F6);
    actionToggleCommentLine->setShortcut(Qt::Key_F7);

    addAction(actionToggleRefineStateDown);
    addAction(actionToggleRefineStateUp);
    addAction(actionToggleCommentLine);

    // connections
    connect(actionSetRefinableIso, SIGNAL(triggered()), this, SLOT(setRefinableIso()));
    connect(actionSetRefinableAniso, SIGNAL(triggered()), this, SLOT(setRefinableAniso()));
    connect(actionSetRefinableFix, SIGNAL(triggered()), this, SLOT(setRefinableFix()));
    connect(actionSetRefinableNumbered, SIGNAL(triggered()), this, SLOT(setRefinableNumbered()));
    connect(actionOpenTextFile, SIGNAL(triggered()), this, SLOT(openTextFile()));
    connect(actionOpenGraphFile, SIGNAL(triggered()), this, SLOT(openGraphFile()));

    connect(actionAddStrucOutFile, SIGNAL(triggered()), this, SLOT(addStrucOutFile()));
    connect(actionAddSimpleStrucOutFile, SIGNAL(triggered()), this, SLOT(addSimpleStrucOutFile()));
    connect(actionAddResFcfOutFile, SIGNAL(triggered()), this, SLOT(addResFcfOutFile()));
    connect(actionAddPdbOutFile, SIGNAL(triggered()), this, SLOT(addPdbOutFile()));

    connect(actionToggleRefineStateDown, SIGNAL(triggered()), this, SLOT(toggleRefineStateDown()));
    connect(actionToggleRefineStateUp, SIGNAL(triggered()), this, SLOT(toggleRefineStateUp()));

    connect(actionCommentLine, SIGNAL(triggered(bool)), this, SLOT(commentLine()));
    connect(actionUncommentLine, SIGNAL(triggered(bool)), this, SLOT(uncommentLine()));
    connect(actionToggleCommentLine, SIGNAL(triggered(bool)), this, SLOT(toggleCommentLine()));

    connect(actionSetParameterYes, SIGNAL(triggered(bool)), this, SLOT(toggleParamYes()));
    connect(actionSetParameterNo, SIGNAL(triggered(bool)), this, SLOT(toggleParamNo()));

    connect(this, SIGNAL(cursorPositionChanged()), this, SLOT(broadcastCurrentWord()));

    paramStart = 0;

    paramsFix.append("A");
    paramsFix.append("B");
    paramsFix.append("C");
    paramsFix.append("ALPHA");
    paramsFix.append("BETA");
    paramsFix.append("GAMMA");
    paramsFix.append("B1");
    paramsFix.append("k1");
    paramsFix.append("k2");
    paramsFix.append("TDS");
    paramsFix.append("x");
    paramsFix.append("y");
    paramsFix.append("z");
    paramsFix.append("EPS1");
    paramsFix.append("EPS2");
    paramsFix.append("EPS3");
    paramsFix.append("EPS4");

    paramsIso.append("A");
    paramsIso.append("B");
    paramsIso.append("C");
    paramsIso.append("ALPHA");
    paramsIso.append("BETA");
    paramsIso.append("GAMMA");
    paramsIso.append("GEWICHT");
    paramsIso.append("B1");
    paramsIso.append("k1");
    paramsIso.append("k2");
    paramsIso.append("TDS");
    paramsIso.append("x");
    paramsIso.append("y");
    paramsIso.append("z");

    paramsAniso.append("A");
    paramsAniso.append("B");
    paramsAniso.append("C");
    paramsAniso.append("ALPHA");
    paramsAniso.append("BETA");
    paramsAniso.append("GAMMA");
    paramsAniso.append("GEWICHT");
    paramsAniso.append("B1");
    paramsAniso.append("k2");
    paramsAniso.append("TDS");
    paramsAniso.append("x");
    paramsAniso.append("y");
    paramsAniso.append("z");

    paramsYesNo.append("PROTOKOLL");
    paramsYesNo.append("SAVE");
    paramsYesNo.append("DDM");

    paramsNumbered.append("EPS1");
    paramsNumbered.append("EPS2");
    paramsNumbered.append("EPS3");
    paramsNumbered.append("EPS4");

    paramsIntegerValue.append("WMIN");
    paramsIntegerValue.append("WMAX");
    paramsIntegerValue.append("RU");
    paramsIntegerValue.append("RP");
    paramsIntegerValue.append("NTHREADS");
    paramsIntegerValue.append("LeBail");

    filesText.append("STRUC");
    filesText.append("LIST");
    filesText.append("OUTPUT");
    filesText.append("STRUCOUT");
    filesText.append("SimpleSTRUCOUT");
    filesText.append("PDBOUT");
    filesText.append("RESOUT");
    filesText.append("FCFOUT");

    filesGraph.append("VAL");
    filesGraph.append("DIAGRAMM");
    filesGraph.append("UNT");
    filesGraph.append("UNTC");
}

void ControlFileEdit::contextMenuEvent(QContextMenuEvent *event)
{
    QMenu *menu = createStandardContextMenu();

    if (parseParam()) {
        menu->addSeparator();

        // add actions if the parameters are refinable
        if (paramsIso.contains(param))      menu->addAction(actionSetRefinableIso);
        if (paramsAniso.contains(param))    menu->addAction(actionSetRefinableAniso);
        if (paramsNumbered.contains(param)) menu->addAction(actionSetRefinableNumbered);
        if (paramsFix.contains(param))      menu->addAction(actionSetRefinableFix);
        if (filesText.contains(param))      menu->addAction(actionOpenTextFile);

        if (filesGraph.contains(param)) {
            menu->addAction(actionOpenTextFile);
            menu->addAction(actionOpenGraphFile);
        }

        if (param == "STRUC") {
            menu->addSeparator();
            menu->addAction(actionAddStrucOutFile);
            menu->addAction(actionAddSimpleStrucOutFile);
            menu->addAction(actionAddResFcfOutFile);
            menu->addAction(actionAddPdbOutFile);
        }

        if (param == "LAMBDA") {
            menu->addSeparator();

            QMultiMapIterator<QString, QString> it(lamFiles);

            while (it.hasNext()) {
                it.next();
                menu->addAction(it.key(), this, SLOT(setLambda()));
            }

            menu->addSeparator();
        }

        if (paramsYesNo.contains(param)) {
            if (isYes()) {
                menu->addAction(actionSetParameterNo);
            } else {
                menu->addAction(actionSetParameterYes);
            }
        }

        if (paramsIntegerValue.contains(param)) {
            menu->addSeparator();
            menu->addAction(actionToggleRefineStateDown);
            menu->addAction(actionToggleRefineStateUp);
        }
    }

    if (isCommented()) {
        menu->addAction(actionUncommentLine);
    } else {
        menu->addAction(actionCommentLine);
    }

    menu->exec(event->globalPos());
    delete menu;
}

void ControlFileEdit::wheelEvent(QWheelEvent *e)
{
    if (e->modifiers() == Qt::ControlModifier) {
        if (e->angleDelta().y() > 0) zoomIn(1);
        else                zoomOut(1);
    } else {
        QPlainTextEdit::wheelEvent(e);
    }
}

void ControlFileEdit::toggleRefineStateDown()
{
    if (!parseParam()) return;

    if (paramsIntegerValue.contains(param)) {
        int n = value.toInt();

        if (param == "RU") n = n <= 1 ? 1 : n - 1;
        if (param == "RP") n = n <= 2 ? 2 : n - 1;
        if (param == "WMIN") n = n <= 1 ? 1 : n - 1;
        if (param == "WMAX") n = n <= 1 ? 1 : n - 1;
        if (param == "NTHREADS") n = n <= 1 ? 1 : n - 1;
        if (param == "LeBail") n = 0;

        replaceCurrentParameter(QString("%1=%2").arg(param).arg(n));
        return;
    }

    if (paramsYesNo.contains(param)) {
        replaceCurrentParameter(QString("%1=N").arg(param));
        return;
    }

    int i = currentRefinementState();

    if (i > 1) {
        if (paramsIso.contains(param)) setRefinableIso();
        return;
    }

    if (i == 1) {
        if (paramsFix.contains(param)) setRefinableFix();
        return;
    }
}

void ControlFileEdit::toggleRefineStateUp()
{
    if (!parseParam()) return;

    if (paramsIntegerValue.contains(param)) {
        int n = value.toInt();

        if (param == "RU") ++n;
        if (param == "RP") n = n >= 4 ? 4 : n + 1;
        if (param == "WMIN") ++n;
        if (param == "WMAX") ++n;
        if (param == "NTHREADS") ++n;
        if (param == "LeBail") n = 1;

        if (param == "RP") {
            replaceCurrentParameter(addRpConditionalParameters(n));
        } else {
            replaceCurrentParameter(QString("%1=%2").arg(param).arg(n));
        }
        return;
    }

    if (paramsYesNo.contains(param)) {
        replaceCurrentParameter(QString("%1=Y").arg(param));
        return;
    }

    int i = currentRefinementState();

    if (i == 0) {
        if (paramsIso.contains(param))      setRefinableIso();
        if (paramsNumbered.contains(param)) setRefinableNumbered();
        return;
    }

    if (i < 3) {
        if (paramsAniso.contains(param)) setRefinableAniso();
        return;
    }
}

QString ControlFileEdit::addRpConditionalParameters(int n)
{
    QString out = QString("RP=%1").arg(n);

    if (n == 2) return out;

    QRegularExpression rxK1("(?:\\s|PARAM=)?k1=\\d+");
    QRegularExpression rxK2("(?:\\s|PARAM=)?k2=(?:ANISO|\\d+)");
    QRegularExpression rxB1("(?:\\s|PARAM=)?B1=(?:ANISO|\\d+)");

    QRegularExpressionMatch rm = rxK1.match(toPlainText());
    bool hasK1 = rm.hasMatch();

    rm = rxK2.match(toPlainText());
    bool hasK2 = rm.hasMatch();

    rm = rxB1.match(toPlainText());
    bool hasB1 = rm.hasMatch();

    if (n == 3) {
        if (!hasB1) out += QString(" PARAM=B1=0_0^0.01");
        return out;
    }

    if (!hasK1) out += QString(" k1=0");
    if (!hasK2) out += QString(" k2=0");
    if (!hasB1) out += QString(" PARAM=B1=0_0^0.01");
    return out;
}

int ControlFileEdit::currentRefinementState()
{
    if (value.left(5) == "ANISO")  return 3; // captures all ANISOxxx

    // these need special treatment, just return 2
    if (value.left(5) == "SPHAR")  return 2;
    if (param == "NTHREADS")       return 2;
    if (param == "PROTOKOLL")      return 2;
    if (param == "SAVE")           return 2;
    if (param == "DDM")            return 2;

    if (prefix.left(5) == "PARAM") {
        if (upper.isEmpty()) return 2; // isotropic but without limits
        else                 return 1; // all normal isotropic ones
    }

    return 0;
}

void ControlFileEdit::toggleParamYes()
{
    replaceCurrentParameter(QString("%1=Y").arg(param));
    resetParamVariables();
}

void ControlFileEdit::toggleParamNo()
{
    replaceCurrentParameter(QString("%1=N").arg(param));
    resetParamVariables();
}

bool ControlFileEdit::parseParam()
{
    // clear old content
    resetParamVariables();

    QTextCursor cursor = textCursor();
    selectedWord = blockUnderSelection();

    // parse the word under the cursor or within the selection
    if (cursor.hasSelection()) {
        splitSelectionUnderCursor(selectedWord, param, value);
    } else {
        splitParameterBlock(selectedWord, prefix, param, value, lower, upper);
    }

    // cut off any "[d]=" at the end of param
    QRegularExpression rxParam("([^\\[\\]=]+)(?:\\[\\d\\])?=?");
    QRegularExpressionMatch rmParam = rxParam.match(param);

    if (!rmParam.hasMatch()) return false;

    param = rmParam.captured(1).trimmed();
    return true;
}

QString ControlFileEdit::lineUnderCursor()
{
    int start = currentLineStart();
    int end   = currentLineEnd();

    return toPlainText().mid(start, end - start);
}

int ControlFileEdit::currentLineStart()
{
    int start = toPlainText().left(textCursor().selectionStart()).lastIndexOf(QRegularExpression("\\n"));
    start = start < 0 ? 0 : start + 1;
    return start;
}

int ControlFileEdit::currentLineEnd()
{
    int end = toPlainText().indexOf(QRegularExpression("\\n|$"), textCursor().selectionEnd());
    return end;
}

QString ControlFileEdit::blockUnderSelection()
{
    QRegularExpression delim("\\s");

    // find the last space before the selection
    int start = toPlainText().left(textCursor().selectionStart()).lastIndexOf(delim);
    int end = toPlainText().indexOf(delim, textCursor().selectionEnd());

    // if a "\\s" was found, we have to cut it off
    paramStart = start < 0 ? 0 : start + 1;

    if (end < 0) return QString();

    return toPlainText().mid(paramStart, end - paramStart);
}

QString ControlFileEdit::wordUnderSelection()
{
    QRegularExpression delim("\\W");

    // find the last space before the selection
    int start = toPlainText().left(textCursor().selectionStart()).lastIndexOf(delim);
    int end = toPlainText().indexOf(delim, textCursor().selectionEnd());

    // if a "\\W" was found, we have to cut it off
    paramStart = start < 0 ? 0 : start + 1;

    if (end < 0) return QString();

    return toPlainText().mid(paramStart, end - paramStart);
}

bool ControlFileEdit::splitParameterBlock(const QString &w, QString &pref, QString &par, QString &val, QString &low, QString &up)
{
    QRegularExpression rxFile(QString("(%1|%2)(?:\\[\\d+\\])?=([\\S_\\.]+)").arg(filesText.join("|")).arg(filesGraph.join("|")));
    QRegularExpression rxParam("(PARAM\\[?\\d*\\]?=)?([a-zA-Z\\d]+\\[?\\d*\\]?)=(-?[^_\\^]+)(_-?\\d+\\.?\\d*)?(\\^-?\\d+\\.?\\d*)?");
    QRegularExpressionMatch match;

    // check if it is a file name
    match = rxFile.match(w);
    if (match.hasMatch()) {
        pref = QString();
        par  = match.captured(1);
        val  = match.captured(2);
        low  = QString();
        up   = QString();
        return true;
    }

    // check if it is a parameter
    match = rxParam.match(w);
    if (match.hasMatch()) {
        pref = match.captured(1);
        par  = match.captured(2);
        val  = match.captured(3);
        low  = match.captured(4).right(match.captured(4).length()-1); // cut off the _ sign
        up   = match.captured(5).right(match.captured(5).length()-1); // cut off the ^ sign
        return true;
    }

    return false;
}

bool ControlFileEdit::splitSelectionUnderCursor(const QString &w, QString &par, QString &val)
{
    QRegularExpression rx("([a-zA-Z\\d]+)(?:\\[\\d+\\])?=(.*)$");
    QRegularExpressionMatch match = rx.match(w);

    if (!match.hasMatch()) {
        qDebug() << QStringLiteral("ControlFileEdit::splitSelectionUnderCursor(): Could not parse word %1").arg(w);
        return false;
    }

    par  = match.captured(1);
    val  = match.captured(2);

    return true;
}

void ControlFileEdit::broadcastCurrentWord()
{
    QStringList lst(absFileName);
    QString word = wordUnderSelection();

    if (absFileName.right(3).toLower() == "sav") {
        if (word == "GOAL")           word = QString("GOAL[i]");
    }

    lst << word;
    emit wordUnderCursor(lst);
}

int ControlFileEdit::fileNumber(const QString &selection)
{
    QRegularExpression rx("[A-Za-z]+\\[(\\d+)\\]=?");
    QRegularExpressionMatch match = rx.match(selection);

    if (match.hasMatch()) {
        return match.captured(1).toInt();
    }

    return -1;
}

QString ControlFileEdit::fileBaseName(const QString &selection)
{
    QRegularExpression rx("[A-Za-z]+\\[\\d+\\]=(.+)\\.(?:str|STR)");
    QRegularExpressionMatch match = rx.match(selection);

    if (match.hasMatch()) {
        return match.captured(1);
    }

    return QString();
}

void ControlFileEdit::setRefinableIso()
{
    double limits = settings->value("cifimport/limits", 0.01).toDouble();

    // Store the value in a double. We may need it as a double value to calculate
    // upper and lower limits relative to the value.
    bool ok;
    double dvalue = value.toDouble(&ok);

    if (!ok) {
        dvalue = 0.0;
    }

    // construct the new parameter string
    QString ostring;

    // reduce SPHAR one step
    if (param == "GEWICHT") {
        if (value.left(5) == "SPHAR") {
            if (value.toUpper() == "SPHAR0") return;
            if (value.toUpper() == "SPHAR2") ostring = QStringLiteral("GEWICHT=SPHAR0");
            if (value.toUpper() == "SPHAR4") ostring = QStringLiteral("GEWICHT=SPHAR2");
            if (value.toUpper() == "SPHAR6") ostring = QStringLiteral("GEWICHT=SPHAR4");
            if (value.toUpper() == "SPHAR8") ostring = QStringLiteral("GEWICHT=SPHAR6");
            if (value.toUpper() == "SPHAR10") ostring = QStringLiteral("GEWICHT=SPHAR8");
        } else {
            ostring = QStringLiteral("GEWICHT=SPHAR2");
        }
    }

    // reduce NTHREADS one step
    if (param.toUpper() == "NTHREADS") {
        int n = value.toInt();
        if (n > 1) ostring = QString("NTHREADS=%1").arg(n - 1);
        else       ostring = QStringLiteral("NTHREADS=1");
    }

    // toggle Protokoll, Save, DDM
    if (param.toUpper() == "PROTOKOLL") {
        if ((value.toUpper() == "Y") || value.toUpper() == "J") ostring = QStringLiteral("PROTOKOLL=N");
        else return;
    }

    if (param.toUpper() == "SAVE") {
        if ((value.toUpper() == "Y") || value.toUpper() == "J") ostring = QStringLiteral("SAVE=N");
        else return;
    }

    if (param.toUpper() == "DDM") {
        if ((value.toUpper() == "Y") || value.toUpper() == "J") ostring = QStringLiteral("DDM=N");
        else return;
    }

    // unit cell dimensions, upper and lower limit of limits*dvalue
    if (param == "A" || param == "B" || param == "C" ||
        param == "ALPHA" || param == "BETA" || param == "GAMMA") {
        ostring = QString("PARAM=%1=%2_%3^%4").arg(param).arg(value).arg(dvalue - dvalue * limits).arg(dvalue + dvalue * limits);
    }

    // k1
    if (param == "k1") {
        ostring = QString("PARAM=k1=%1_0^1").arg(value);
    }

    // k2
    if (param == "k2") {
        if (upper.isEmpty()) upper = QStringLiteral("0.0001");
        ostring = QString("PARAM=k2=%1_0^%2").arg(dvalue).arg(upper);
    }

    // B1
    if (param == "B1") {
        if (upper.isEmpty()) upper = QStringLiteral("0.01");
        ostring = QString("PARAM=B1=%1_0^%2").arg(dvalue).arg(upper);
    }

    // tds, lower limit = 0, upper limit = 2*tds
    if (param == "TDS") {
        if (upper.isEmpty()) {
            // fallback value
            upper = QStringLiteral("0.02");
            // fallback 2, in case the value is greater than the fallback upper limit
            if (dvalue > 0.02) upper = QString("%1").arg(2.0 * dvalue, 0, 'f', 6);
        }
        ostring = QString("PARAM=TDS=%1_0^%2").arg(dvalue, 0, 'f', 6).arg(upper);
    }

    // fractional coordinates, upper and lower limit of delta
    if (param == "x" || param == "y" || param == "z") {
        double fcoordLimit = settings->value("bgmnProject/coordinatelimits", 0.05).toDouble();
        ostring = QString("PARAM=%1=%2_%3^%4").arg(param).arg(value).arg(dvalue - fcoordLimit, 0, 'f', 4).arg(dvalue + fcoordLimit, 0, 'f', 4);
    }

    replaceCurrentParameter(ostring);
    resetParamVariables();
}

void ControlFileEdit::setRefinableAniso()
{
    // construct the new parameter string
    QString ostring;

    if (param == "GEWICHT") {
        if (value.left(5) == "SPHAR") {
            if (value.toUpper() == "SPHAR10") return;
            if (value.toUpper() == "SPHAR8") ostring = QStringLiteral("GEWICHT=SPHAR10");
            if (value.toUpper() == "SPHAR6") ostring = QStringLiteral("GEWICHT=SPHAR8");
            if (value.toUpper() == "SPHAR4") ostring = QStringLiteral("GEWICHT=SPHAR6");
            if (value.toUpper() == "SPHAR2") ostring = QStringLiteral("GEWICHT=SPHAR4");
            if (value.toUpper() == "SPHAR0") ostring = QStringLiteral("GEWICHT=SPHAR2");
        } else {
            ostring = QStringLiteral("GEWICHT=SPHAR2");
        }
    }

    // increase NTHREADS one step
    if (param.toUpper() == "NTHREADS") {
        int n = value.toInt();
        if (n < maxNThreads) ostring = QString("NTHREADS=%1").arg(n + 1);
        else                 ostring = QString("NTHREADS=%1").arg(maxNThreads);
    }

    // toggle Protokoll, Save, DDM
    if (param.toUpper() == "PROTOKOLL") {
        if (value.toUpper() == "N") ostring = QStringLiteral("PROTOKOLL=Y");
        else return;
    }

    if (param.toUpper() == "SAVE") {
        if (value.toUpper() == "N") ostring = QStringLiteral("SAVE=Y");
        else return;
    }

    if (param.toUpper() == "DDM") {
        if (value.toUpper() == "N") ostring = QStringLiteral("DDM=Y");
        else return;
    }

    // k2
    if (param == "k2") {
        if (upper.isEmpty()) {
            ostring = QStringLiteral("k2=ANISO4^0.0001");
        } else {
            ostring = QString("k2=ANISO4^%1").arg(upper);
        }
    }

    // B1
    if (param == "B1") {
        if (upper.isEmpty()) {
            ostring = QStringLiteral("B1=ANISO^0.01");
        } else {
            ostring = QString("B1=ANISO^%1").arg(upper);
        }
    }

    // TDS
    if (param == "TDS") {
        if (upper.isEmpty()) {
            ostring = QStringLiteral("TDS=ANISO^0.02");
        } else {
            ostring = QString("TDS=ANISO^%1").arg(upper);
        }
    }

    // fractional coordinates, not really anisotropic, but refined without limits
    if (param == "x" || param == "y" || param == "z") {
        ostring = QString("PARAM=%1=%2").arg(param).arg(value);
    }

    // unit cell, not really anisotropic, but refined without limits
    if (param == "A" || param == "B" || param == "C"
            || param == "ALPHA" || param == "BETA" || param == "GAMMA") {
        ostring = QString("PARAM=%1=%2").arg(param).arg(value);
    }

    replaceCurrentParameter(ostring);
    resetParamVariables();
}

void ControlFileEdit::setRefinableFix()
{
    // Store the value in a double. If value is something like ANISO, the conversion
    // will fail and we will fall back to 0.0, which is the intended behaviour
    bool ok;
    double dvalue = value.toDouble(&ok);

    if (!ok) {
        dvalue = 0.0;
    }

    // construct the new parameter string
    QString ostring;

    // special replacements

    // GEWICHT hard coded to SPHAR0
    if (param == "GEWICHT") {
        ostring = QStringLiteral("GEWICHT=0.1");
    }

    // unit cell dimensions, upper and lower limit of 0.01*dvalue
    if (param == "A" || param == "B" || param == "C" ||
        param == "ALPHA" || param == "BETA" || param == "GAMMA") {
        ostring = QString("%1=%2").arg(param).arg(value);
    }

    // k1
    if (param == "k1") {
        ostring = QString("k1=%1").arg(value);
    }

    // k2
    if (param == "k2") {
        ostring = QString("k2=%1").arg(dvalue);
    }

    // B1
    if (param == "B1") {
        ostring = QString("B1=%1").arg(dvalue);
    }

    // TDS
    if (param == "TDS") {
        ostring = QString("TDS=%1").arg(dvalue);
    }

    // fractional coordinates
    if (param == "x" || param == "y" || param == "z") {
        ostring = QString("%1=%2").arg(param).arg(value);
    }

    // EPS values
    if (param.left(3) == "EPS") {
        ostring = QString("%1=%2").arg(param).arg(value);
    }

    replaceCurrentParameter(ostring);

    if (paramsNumbered.contains(param)) {
        fixParameterNumbering("PARAM");
    }

    resetParamVariables();
}

void ControlFileEdit::setRefinableNumbered()
{
    // construct the new parameter string
    QString ostring;

    int n = freeParamNumber();

    // EPS values
    if (param.left(3) == "EPS") {
        ostring = QString("PARAM[%1]=%2=%3_-0.01^0.01").arg(n).arg(param).arg(value);
    }

    replaceCurrentParameter(ostring);
    fixParameterNumbering("PARAM");
    resetParamVariables();
}

void ControlFileEdit::replaceCurrentParameter(const QString &s)
{
    // select and clear the text
    QTextCursor cursor = textCursor();
    cursor.setPosition(paramStart);
    cursor.movePosition(QTextCursor::Right, QTextCursor::KeepAnchor, selectedWord.length());
    cursor.removeSelectedText();

    // write back the string
    cursor.insertText(s);
    cursor.setPosition(paramStart);
    setTextCursor(cursor);
}

void ControlFileEdit::openTextFile()
{
    emit openText(value.trimmed());
}

void ControlFileEdit::openGraphFile()
{
    emit openGraph(value.trimmed());
}

void ControlFileEdit::addStrucOutFile()
{
    QString line = blockUnderSelection();
    int n = fileNumber(line);
    QString bn = fileBaseName(line);

    if ((n >= 0) && !bn.isNull()) {
        QString newLine = QString("STRUCOUT[%1]=%2-%3.str").arg(n).arg(bn).arg(editorBaseName);
        insertLine(line, newLine, false);
    }
}

void ControlFileEdit::addSimpleStrucOutFile()
{
    QString line = blockUnderSelection();
    int n = fileNumber(line);
    QString bn = fileBaseName(line);

    if ((n >= 0) && !bn.isNull()) {
        QString newLine = QString("SimpleSTRUCOUT[%1]=%2-simple-%3.str").arg(n).arg(bn).arg(editorBaseName);
        insertLine(line, newLine, false);
    }
}

void ControlFileEdit::addResFcfOutFile()
{
    QString line = blockUnderSelection();
    int n = fileNumber(line);
    QString bn = fileBaseName(line);

    if ((n >= 0) && !bn.isNull()) {
        QString newResLine = QString("RESOUT[%1]=%2-%3.res").arg(n).arg(bn).arg(editorBaseName);
        QString newFcfLine = QString("FCFOUT[%1]=%2-%3.fcf").arg(n).arg(bn).arg(editorBaseName);

        insertLine(line, newResLine, false);
        insertLine(line, newFcfLine, false);
    }
}

void ControlFileEdit::addPdbOutFile()
{
    QString line = blockUnderSelection();
    int n = fileNumber(line);
    QString bn = fileBaseName(line);

    if ((n >= 0) && !bn.isNull()) {
        QString newLine = QString("PDBOUT[%1]=%2-%3.pdb").arg(n).arg(bn).arg(editorBaseName);
        insertLine(line, newLine, false);
    }
}

void ControlFileEdit::insertLine(const QString &before, const QString &newLine, bool allowDuplicates)
{
    QString str = toPlainText();

    if (str.contains(newLine.trimmed())) {
        if (!allowDuplicates) {
            qDebug() << QString("ControlFileEdit::insertLine(): Line is already present.").arg(newLine);
            return;
        }
    }

    QStringList content = str.split(global::rxLineEnding);

    for (int i = 0; i < content.size(); ++i) {
        if (content.at(i).contains(before.trimmed())) {
            content.insert(i+1, newLine.trimmed());
            break;
        }
    }

    setPlainText(content.join("\n"));
}

void ControlFileEdit::replaceCurrentLine(const QString &newLine)
{
    QTextCursor cursor = textCursor();
    int cpos = textCursor().position();
    QString str = toPlainText();

    int lineStart = currentLineStart();
    int lineEnd   = currentLineEnd();

    str.replace(lineStart, lineEnd - lineStart, newLine);

    setPlainText(str);
    cursor.setPosition(cpos);
    setTextCursor(cursor);
}

void ControlFileEdit::resetParamVariables()
{
    selectedWord = QString();
    prefix = QString();
    param = QString();
    value = QString();
    lower = QString();
    upper = QString();
    paramStart = -1;
}

int ControlFileEdit::freeParamNumber()
{
    QRegularExpression rx("\\bPARAM\\[(\\d+)\\]=");
    QString content = toPlainText();

    // if no PARAM[n]= line is found, we can use number 1
    if (!content.contains(rx)) return 1;

    // loop over all PARAM[n]= lines and store n in a list
    QList<int> lst;
    int i = content.indexOf(rx, 0);
    int max = 0;

    while (i >= 0) {
        QRegularExpressionMatch rm = rx.match(content, i);
        if (rm.hasMatch()) {
            int v = rm.captured(1).toInt();
            max = qMax(v, max);
            lst.append(v);
        }

        i = content.indexOf(rx, i + 1);
    }

    // try to find the first gap in the list
    for (int n = 1; n < max; ++n) {
        if (!lst.contains(n)) return n;
    }

    // if no gap was found, return max + 1
    return max + 1;
}

void ControlFileEdit::fixAllNumbering()
{
    fixParameterNumbering("PARAM");
    fixParameterNumbering("STRUC");
    fixParameterNumbering("VAL");
    fixParameterNumbering("CUT");
}

void ControlFileEdit::fixParameterNumbering(const QString &pattern)
{
    QTextCursor cursor = textCursor();
    int cpos = textCursor().position();

    QString content = toPlainText();
    QRegularExpression rx(QString("(%\\s*)?(%1\\[\\d+\\]=)").arg(pattern));
    QRegularExpressionMatchIterator ri = rx.globalMatch(content);

    int n = 1;

    while (ri.hasNext()) {
        QRegularExpressionMatch rm = ri.next();

        if (rm.captured(1).isEmpty()) {
            content.replace(rm.capturedStart(2), rm.capturedLength(2), QString("%1[%2]=").arg(pattern).arg(n));
            ++n;
        }
    }

    setPlainText(content);
    cursor.setPosition(cpos);
    setTextCursor(cursor);
}

int ControlFileEdit::replaceAll(QString findString, bool isExpr, QString replaceString, bool caseSensitively, bool wholeWords)
{
    QTextDocument *doc = document();
    QTextCursor cursor = textCursor();
    cursor.beginEditBlock();
    cursor.movePosition(QTextCursor::Start);
    QTextCursor newCursor = cursor;
    int count = 0;

    QTextDocument::FindFlags options;
    if (caseSensitively) options = options | QTextDocument::FindCaseSensitively;
    if (wholeWords)      options = options | QTextDocument::FindWholeWords;

    if (!findString.isEmpty())
    {
        while (true) {
            if (isExpr) {
                newCursor = doc->find(QRegularExpression(findString), newCursor, options);
            } else {
                newCursor = doc->find(findString, newCursor, options);
            }

            if (!newCursor.isNull()) {
                if (newCursor.hasSelection()) {
                    newCursor.insertText(replaceString);
                    count++;
                }
            } else {
                break;
            }
        }
    }

    cursor.endEditBlock();
    return count;
}

void ControlFileEdit::commentLine()
{
    QString c(commentSign());
    QString s(lineUnderCursor());

    if (s.trimmed().left(c.length()) != c) {
        replaceCurrentLine(QString("%1 %2").arg(c).arg(s));
    }

    fixAllNumbering();
}

void ControlFileEdit::uncommentLine()
{
    QString s(lineUnderCursor());
    QRegularExpression rx("(?:[\\/|%]+\\s*)?(.*)");
    QRegularExpressionMatch rm = rx.match(s);

    if (rm.hasMatch()) {
        replaceCurrentLine(rm.captured(1));
    }

    fixAllNumbering();
}

bool ControlFileEdit::isCommented()
{
    QString s(lineUnderCursor());
    QRegularExpression rx("^\\s*[\\/|%]+\\s*.*");
    QRegularExpressionMatch rm = rx.match(s);
    return rm.hasMatch();
}

void ControlFileEdit::toggleCommentLine()
{
    if (isCommented()) {
        uncommentLine();
    } else {
        commentLine();
    }
}

QString ControlFileEdit::commentSign()
{
    QString s = toPlainText();

    if (s.contains("PHASE="))   return QString("//");
    if (s.contains("VERZERR=")) return QString("%");

    // fallback
    return QString("%");
}

bool ControlFileEdit::isYes()
{
    if (value == "Y") return true;
    if (value == "y") return true;
    if (value == "J") return true;
    if (value == "j") return true;
    return false;
}

void ControlFileEdit::setLambda()
{
    QString lam = static_cast<QAction*>(sender())->text();
    qDebug() << QString("ControlFileEdit::setLambda(): lam file selected: %1").arg(lam);

    QRegularExpression rx("LAMBDA=\\S+");

    QString s = toPlainText();
    s.replace(rx, QString("LAMBDA=%1").arg(lam));
    setPlainText(s);
}

QMap<int, QStringList> ControlFileEdit::findAll(const QString &pattern)
{
    QRegularExpression rx(pattern);
    QStringList content = toPlainText().split("\n");

    QMap<int, QStringList> matches;

    for (int i = 0; i < content.size(); ++i) {
        QRegularExpressionMatch rm = rx.match(content.at(i));
        if (rm.hasMatch()) matches[i+1] = QStringList() << rm.captured(1) << content.at(i);
    }

    return matches;
}

void ControlFileEdit::setCurrentLine(int l)
{
    //QTextCursor cur = textCursor();
    //cur.movePosition(QTextCursor::Start);
    //cur.movePosition(QTextCursor::Down, QTextCursor::MoveAnchor, l);
    moveCursor(QTextCursor::End);
    QTextCursor cur(document()->findBlockByLineNumber(l));
    setTextCursor(cur);
}

