/***************************************************************************
                          graphdatacontroller.cpp  -  description
                             -------------------
    begin                : Tue Feb 11 18:00:00 CEST 2020
    copyright            : (C) 2020 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "graphdatacontroller.h"
#include <limits>

GraphDataController::GraphDataController(GraphDataModel *m, QObject *parent) :
    graphModel(m),
    QObject(parent)
{
}

void GraphDataController::addView(AbstractGraphView *v)
{
    graphViews.append(v);
}

int GraphDataController::count() const
{
    if (!graphModel) return 0;
    return graphModel->count();
}

int GraphDataController::countFiles() const
{
    if (!graphModel) return 0;
    return graphModel->numberOfFiles();
}

int GraphDataController::loadScanFile(const QString &s, const QString &uid, bool min, const QString &sid, bool upd)
{
    if (!graphModel) return -1;

    int n = graphModel->loadScanFile(s, uid, min, sid);
    if (upd) {
        resetViews();
        updateViews();
    }

    return n;
}

int GraphDataController::loadDiaFile(const QString &s, bool min, const QString &sid, bool upd, int &ll, int &tl)
{
    if (!graphModel) return -1;

    int n = graphModel->loadDiaFile(s, min, sid, ll, tl);
    if (upd) {
        resetViews();
        updateViews();
    }

    return n;
}

bool GraphDataController::saveScanFile(const QString &f, const QString &uid, int n, const QMap<QString, QVariant> &flags)
{
    if (!graphModel) return false;
    return graphModel->saveScanFile(f, uid, n, flags);
}

int GraphDataController::addScanFile(const QString &s, const QString &uid, bool min, bool upd)
{
    if (!graphModel) return -1;

    int n = graphModel->addScanFile(s, uid, min);
   if (upd) updateViews();

    return n;
}

void GraphDataController::appendScan(const Scan &s, bool upd)
{
    if (!graphModel) return;

    graphModel->appendScan(s);
    if (upd) updateViews();
}

void GraphDataController::appendHklScan(const Scan &s, bool upd)
{
    if (!graphModel) return;

    graphModel->appendHklScan(s);
    if (upd) updateViews();
}

void GraphDataController::replaceScan(int n, const Scan &s, bool upd)
{
    if (!graphModel) return;

    graphModel->replaceScan(n, s);
    if (upd) updateViews();
}

void GraphDataController::replaceHklScan(int n, const Scan &s, bool upd)
{
    if (!graphModel) return;

    graphModel->replaceHklScan(n, s);
    if (upd) updateViews();
}

void GraphDataController::setAllHklDisplayStatus(int i, bool upd)
{
    if (!graphModel) return;

    graphModel->setAllHklDisplayStatus(i);
    if (upd) updateViews();
}

void GraphDataController::removeScan(int n, bool upd)
{
    if (!graphModel) return;
    graphModel->removeScan(n);
    if (upd) updateViews();
}

void GraphDataController::removeScan(const QUuid &u, bool upd)
{
    if (!graphModel) return;
    graphModel->removeScan(u);
    if (upd) updateViews();
}

int GraphDataController::reloadScanFile(bool min, bool upd)
{
    int n = -1;

    if (!graphModel) return n;

    n = graphModel->reloadScanFile(min);
    if (upd) updateViews();
    return n;
}

int GraphDataController::reloadDiaFile(bool min, bool upd, int &ll, int &tl)
{
    int n = -1;

    if (!graphModel) return n;

    n = graphModel->reloadDiaFile(min, ll, tl);
    if (upd) updateViews();
    return n;
}

/*
 * the sender will be skipped. set it to nullptr if the sender should be updated too
 */
void GraphDataController::updateViews(const AbstractGraphView *sender)
{
    for (int i = 0; i < graphViews.size(); ++i) {
        if (sender != graphViews.at(i))
            graphViews.at(i)->updateView();
    }

    emit dataUpdated();
}

/*
 * the sender will be skipped. set it to nullptr if the sender should be updated too
 */
void GraphDataController::resetViews(const AbstractGraphView *sender)
{
    for (int i = 0; i < graphViews.size(); ++i) {
        if (sender != graphViews.at(i))
            graphViews.at(i)->resetView();
    }
}

void GraphDataController::setViewStatus(const global::RefinementStatus &n)
{
    for (int i = 0; i < graphViews.size(); ++i) {
        if (graphViews.at(i)->setStatus(n)) {
            graphViews.at(i)->updateView();
        }
    }
}

QString GraphDataController::fileName() const
{
    if (!graphModel) return QString();
    return graphModel->fileInfo().absoluteFilePath();
}

QFileInfo GraphDataController::fileInfo() const
{
    if (!graphModel) return QFileInfo();
    return graphModel->fileInfo();
}

bool GraphDataController::hasData() const
{
    if (!graphModel) return false;
    return graphModel->hasData();
}

bool GraphDataController::hasScanData() const
{
    if (!graphModel) return false;
    return graphModel->hasScanData();
}

bool GraphDataController::hasHklData() const
{
    if (!graphModel) return false;
    return graphModel->hasHklData();
}

QList<int> GraphDataController::baseLines() const
{
    if (!graphModel) return QList<int>();
    return graphModel->baseLines();
}

double GraphDataController::getWaveLength(double fallback)
{
    if (!graphModel) return fallback;
    return graphModel->getWaveLength(fallback);
}

double GraphDataController::getTimePerStep(int i)
{
    if (!graphModel) return -1.0;
    if (i >= graphModel->count()) return -1.0;
    return graphModel->at(i)->timePerStep();
}

Scan * GraphDataController::getScan(int i)
{
    if (!graphModel) return nullptr;
    return graphModel->getScan(i);
}

Scan * GraphDataController::getScan(const QUuid &u)
{
    if (!graphModel) return nullptr;
    return graphModel->getScan(u);
}

Scan * GraphDataController::getScan(const QString &s)
{
    if (!graphModel) return nullptr;

    for (int i = 0; i < graphModel->count(); ++i) {
        if (graphModel->at(i)->name() == s) {
            return graphModel->getScan(i);
        }
    }

    return nullptr;
}

QStringList GraphDataController::getAllScanNames() const
{
    if (!graphModel) return QStringList();

    QStringList l;
    const QVector<const Scan *> v = graphModel->allScans();

    for (int i = 0; i < v.size(); ++i) {
        l.append(v.at(i)->name());
    }

    return l;
}

QStringList GraphDataController::getXyScanNames() const
{
    if (!graphModel) return QStringList();

    QStringList l;
    const QVector<const Scan *> v = graphModel->xyScans();

    for (int i = 0; i < v.size(); ++i) {
        l.append(v.at(i)->name());
    }

    return l;
}

QStringList GraphDataController::getHklScanNames() const
{
    if (!graphModel) return QStringList();

    QStringList l;
    const QVector<const Scan *> v = graphModel->hklScans();

    for (int i = 0; i < v.size(); ++i) {
        l.append(v.at(i)->name());
    }

    return l;
}

Scan * GraphDataController::getFirst()
{
    if (!graphModel) return nullptr;
    return graphModel->getFirst();
}

Scan * GraphDataController::getLast()
{
    if (!graphModel) return nullptr;
    return graphModel->getLast();
}

const Scan * GraphDataController::first()
{
    if (!graphModel) return nullptr;
    return graphModel->first();
}

const Scan * GraphDataController::last()
{
    if (!graphModel) return nullptr;
    return graphModel->last();
}

const Scan * GraphDataController::at(int i)
{
    if (!graphModel) return nullptr;
    return graphModel->at(i);
}

Scan * GraphDataController::getFirstXy()
{
    if (!graphModel) return nullptr;
    return graphModel->getFirstXy();
}

Scan * GraphDataController::getFirstHkl()
{
    if (!graphModel) return nullptr;
    return graphModel->getFirstHkl();
}

int GraphDataController::indexOf(const Scan &s) const
{
    if (!graphModel) return -1;
    return graphModel->indexOf(s);
}

int GraphDataController::indexOf(const QUuid &u) const
{
    if (!graphModel) return -1;
    return graphModel->indexOf(u);
}

QUuid GraphDataController::scanUid(int i) const
{
    if (!graphModel) return QUuid();
    return graphModel->scanUid(i);
}

void GraphDataController::updateScanColors() const
{
    if (!graphModel) return;
    graphModel->updateScanColors();
}

void GraphDataController::updateScanStyles() const
{
    if (!graphModel) return;
    graphModel->updateScanStyles();
}

void GraphDataController::setAngularCorrections(double e1, double e2, double e3)
{
    if (!graphModel) return;
    graphModel->setAngularCorrections(e1, e2, e3);
}

double GraphDataController::getAngularCorrEPS1() const
{
    if (!graphModel) return 0.0;
    return graphModel->getAngularCorrEPS1();
}

double GraphDataController::getAngularCorrEPS2() const
{
    if (!graphModel) return 0.0;
    return graphModel->getAngularCorrEPS2();
}

double GraphDataController::getAngularCorrEPS3() const
{
    if (!graphModel) return 0.0;
    return graphModel->getAngularCorrEPS3();
}

QString GraphDataController::getSampleId() const
{
    if (!graphModel) return QString();
    return graphModel->getSampleId();
}

bool GraphDataController::getNearestCharacteristicWaveLength(double &ka1, double &ka2, double &kb, double fixwl)
{
    if (graphModel) return graphModel->getNearestCharacteristicWaveLength(ka1, ka2, kb, fixwl);
    return false;
}

void GraphDataController::setCustomWaveLength(const QMap<QString, double> &m)
{
    if (graphModel) graphModel->setCustomWaveLength(m);
}

void GraphDataController::clearCustomWaveLength()
{
    if (graphModel) graphModel->clearCustomWaveLength();
}

const QMap<QString, double> *GraphDataController::getCustomWaveLength()
{
    if (!graphModel) return nullptr;
    return graphModel->getCustomWaveLength();
}

bool GraphDataController::sortScanOrder(const QVector<QUuid> &v, bool upd, const AbstractGraphView *sender)
{
    if (!v.size())   return true;
    if (!graphModel) return false;

    if (graphModel->sortScanOrder(v)) {
        if (upd) updateViews(sender);
        return true;
    }

    return false;
}

double GraphDataController::getXmin() const
{
    if (!graphModel) return 0.0;
    return graphModel->getXmin();
}

double GraphDataController::getXmax() const
{
    if (!graphModel) return 0.0;
    return graphModel->getXmax();
}

double GraphDataController::getYmin() const
{
    if (!graphModel) return 0.0;
    return graphModel->getYmin();
}

double GraphDataController::getYmax() const
{
    if (!graphModel) return 0.0;
    return graphModel->getYmax();
}

void GraphDataController::yOffsetUp()
{
    if (graphModel) graphModel->yOffsetUp();
}

void GraphDataController::yOffsetDown()
{
    if (graphModel) graphModel->yOffsetDown();
}

void GraphDataController::xOffsetRight()
{
    if (graphModel) graphModel->xOffsetRight();
}

void GraphDataController::xOffsetLeft()
{
    if (graphModel) graphModel->xOffsetLeft();
}

void GraphDataController::resetOffset()
{
    if (graphModel) graphModel->resetOffset();
}

bool GraphDataController::isStacked() const
{
    if (!graphModel) return false;
    return graphModel->isStacked();
}

void GraphDataController::keepTemporary()
{
    if (graphModel) graphModel->keepTemporary();
}

void GraphDataController::keepTemporary(const QList<QUuid> &l)
{
    if (graphModel) graphModel->keepTemporary(l);
}

void GraphDataController::clearTemporary()
{
    if (graphModel) graphModel->clearTemporary();
}

void GraphDataController::removeScans(const QList<QUuid> &l)
{
    if (graphModel) graphModel->removeScans(l);
}

void GraphDataController::normalizeHklScan(Scan *s)
{
    if (graphModel) graphModel->normalizeHklScan(s);
}

QVector<Scan *> GraphDataController::activeScans()
{
    if (!graphModel) return QVector<Scan*>();
    return graphModel->activeScans();
}

void GraphDataController::setActiveScans(const QList<int> &l)
{
    if (graphModel) graphModel->setActiveScans(l);
}

void GraphDataController::setActiveScans(const QList<QUuid> &l)
{
    if (graphModel) graphModel->setActiveScans(l);
}

QList<int> GraphDataController::activeScanIncides() const
{
    if (!graphModel) return QList<int>();
    return graphModel->activeScanIndices();
}

bool GraphDataController::hasActiveScan() const
{
    if (!graphModel) return false;
    return graphModel->activeScanIndices().size() > 0;
}

Scan * GraphDataController::firstActiveScan()
{
    if (!graphModel) return nullptr;

    QVector<Scan*> aSc = graphModel->activeScans();
    if (aSc.size()) return aSc.first();

    return graphModel->getFirst();
}

void GraphDataController::clearActiveScans() const
{
    if (!graphModel) return;
    graphModel->clearActiveScans();
}

int GraphDataController::backgroundScanIndex() const
{
    if (!graphModel) return -1;
    return graphModel->backgroundScanIndex();
}

Scan * GraphDataController::backgroundScan()
{
    if (!graphModel) return nullptr;
    return graphModel->getScan(graphModel->backgroundScanIndex());
}

void GraphDataController::setScanVisibility(int n, bool v)
{
    if (!graphModel) return;
    Scan *scan = graphModel->getScan(n);
    if (scan) scan->setVisible(v);
}

void GraphDataController::setPhaseVisibility(bool v)
{
    if (!graphModel) return;

    for (int i = 0; i < graphModel->count(); ++i) {
        Scan *scan = graphModel->getScan(i);
        if (!scan) continue;
        if (scan->scanTypes().testFlag(Scan::PHASE)) {
            scan->setVisible(v);
        }
    }
}

void GraphDataController::togglePhaseVisibility()
{
    if (!graphModel) return;

    QList<Scan*> pScans;

    int vis = 0;
    int nvis = 0;

    // check if there are more visible or non-visible phase scans
    for (int i = 0; i < graphModel->count(); ++i) {
        Scan *scan = graphModel->getScan(i);
        if (!scan) continue;
        if (scan->scanTypes().testFlag(Scan::PHASE)) {
            pScans.append(scan);
            if (scan->isVisible()) vis++;
            else                   nvis++;
        }
    }

    bool v = nvis >= vis;

    for (int i = 0; i < pScans.size(); ++i) {
        pScans[i]->setVisible(v);
    }
}

void GraphDataController::setWavelengthMode(Scan::WavelengthMode m)
{
    graphModel->setWavelengthMode(m);
}

Scan::WavelengthMode GraphDataController::getWavelengthMode() const
{
    return graphModel->getWavelengthMode();
}

const QVector<const Scan *> GraphDataController::xyScans()
{
    if (graphModel) return graphModel->xyScans();
    return QVector<const Scan *>();
}

const QVector<const Scan *> GraphDataController::hklScans()
{
    if (graphModel) return graphModel->hklScans();
    return QVector<const Scan *>();
}

const QVector<const Scan *> GraphDataController::allScans()
{
    if (graphModel) return graphModel->allScans();
    return QVector<const Scan *>();
}

const QVector<const Scan *> GraphDataController::phaseScans()
{
    if (graphModel) return graphModel->phaseScans();
    return QVector<const Scan *>();
}
