/***************************************************************************
                          baselinedialog.cpp  -  description
                             -------------------
    begin                : Tue Jan 23 22:00:00 CEST 2018
    copyright            : (C) 2018 by Nicola Doebelin
    email                : ndoebelin@gmx.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QComboBox>
#include <QSpinBox>
#include <QShowEvent>
#include <QCloseEvent>
#include <QTableWidget>
#include <limits.h>
#include <QDebug>
#include "../libXrdIO/structs.h"
#include "scansmoothdialog.h"
#include "ui_scansmoothdialog.h"

ScanSmoothDialog::ScanSmoothDialog(QWidget *parent) :
    AbstractToolDialog(parent),
    ui(new Ui::ScanSmoothDialog)
{
    ui->setupUi(this);
    setWindowTitle(tr("Smooth Scans"));

    temporaryScan = nullptr;

    initSettings();

    connect(ui->comboBoxAlgo, SIGNAL(currentIndexChanged(int)), this, SLOT(algoChanged(int)));
    connect(ui->comboBoxScan, SIGNAL(currentIndexChanged(int)), this, SLOT(compute()));

    connect(ui->spinBoxSmaWindow,  SIGNAL(valueChanged(int)), this, SLOT(computeSma()));
    connect(ui->spinBoxEmaWindow,  SIGNAL(valueChanged(double)), this, SLOT(computeEma()));
    connect(ui->spinBoxLrmaWindow, SIGNAL(valueChanged(int)), this, SLOT(computeLrma()));
}

ScanSmoothDialog::~ScanSmoothDialog()
{
    delete ui;
}

void ScanSmoothDialog::initSettings()
{
    ui->comboBoxAlgo->setCurrentIndex(settings->value("smoothDialog/algorithm", 0).toInt());
    ui->stackedWidgetParams->setCurrentIndex(ui->comboBoxAlgo->currentIndex());

    ui->spinBoxSmaWindow->setValue(settings->value("smoothDialog/algoSma/window", 5).toInt());
    ui->spinBoxEmaWindow->setValue(settings->value("smoothDialog/algoEma/window", 0.5).toDouble());
    ui->spinBoxLrmaWindow->setValue(settings->value("smoothDialog/algoLrma/window", 5).toInt());

    restoreGeometry(settings->value("smoothDialog/geometry", QByteArray()).toByteArray());
}

void ScanSmoothDialog::saveSettings()
{
    settings->setValue("smoothDialog/algorithm", ui->comboBoxAlgo->currentIndex());

    settings->setValue("smoothDialog/algoSma/window", ui->spinBoxSmaWindow->value());
    settings->setValue("smoothDialog/algoEma/window", ui->spinBoxEmaWindow->value());
    settings->setValue("smoothDialog/algoLrma/window", ui->spinBoxLrmaWindow->value());

    settings->setValue("smoothDialog/geometry", saveGeometry());
}

void ScanSmoothDialog::preSetProject(ProjectWidget *)
{
    temporaryScan = nullptr;
}

void ScanSmoothDialog::postSetProject(ProjectWidget *)
{
    clearTemporary();

    if (!checkBackend()) {
        clearGui();
        return;
    }

    parseScans();
    compute();
}

void ScanSmoothDialog::updateView()
{
    if (!isVisible()) return;
    parseScans();
}

void ScanSmoothDialog::compute()
{
    switch (ui->stackedWidgetParams->currentIndex()) {
        case 0:
            computeAlgo(SMA);
            break;
        case 1:
            computeAlgo(EMA);
            break;
        case 2:
            computeAlgo(LRMA);
    }
}

void ScanSmoothDialog::algoChanged(int i)
{
    ui->stackedWidgetParams->setCurrentIndex(i);
    compute();
}

void ScanSmoothDialog::computeAlgo(SmoothAlgoType a)
{
    if (!checkBackend()) return;

    int c = ui->comboBoxScan->currentIndex();
    if (c < 0) return;

    const Scan *scan = graphControl->getScan(ui->comboBoxScan->currentData(Qt::UserRole).toUuid());
    if (!scan) return;

    Scan sm = scan->clone();
    sm.pDataHkl().clear();

    switch (a) {
        case SMA:
            sm.setDataInt(ScanOps::smoothMovingAverageSimple(scan->pDataIntensity(), ui->spinBoxSmaWindow->value()));
            break;
        case EMA:
            sm.setDataInt(ScanOps::smoothMovingAverageExponential(scan->pDataIntensity(), ui->spinBoxEmaWindow->value()));
            break;
        case LRMA:
            sm.setDataInt(ScanOps::smoothMovingAverageLinearRegression(scan->pDataIntensity(), ui->spinBoxLrmaWindow->value()));
    }

    sm.setName(QString("Smoothed %1").arg(ui->comboBoxScan->currentText()));
    sm.setTypes(Scan::XY | Scan::SYNTHETIC | Scan::TEMPORARY);
    sm.setColor(QString());

    int idx = tempScanIndex();

    if (idx < 0) {
        graphControl->appendScan(sm, true);
        temporaryScan = graphControl->getLast();
    } else {
        graphControl->replaceScan(idx, sm, true);
        temporaryScan = graphControl->getScan(idx);
    }
}

void ScanSmoothDialog::computeSma()
{
    computeAlgo(SMA);
}

void ScanSmoothDialog::computeEma()
{
    computeAlgo(EMA);
}

void ScanSmoothDialog::computeLrma()
{
    computeAlgo(LRMA);
}

void ScanSmoothDialog::parseScans()
{
    bool oldState = ui->comboBoxScan->blockSignals(true);
    int n = ui->comboBoxScan->currentIndex();

    ui->comboBoxScan->clear();

    if (!graphControl) {
        ui->comboBoxScan->blockSignals(oldState);
        return;
    }

    for (int i = 0; i < graphControl->count(); ++i) {
        if (!graphControl->at(i)->hasScanData()) continue;
        if (graphControl->at(i)->isTemporary())  continue;

        ui->comboBoxScan->addItem(QString(graphControl->at(i)->name()), graphControl->at(i)->uid());
    }

    n = n < ui->comboBoxScan->count() ? n : ui->comboBoxScan->count() - 1;
    ui->comboBoxScan->setCurrentIndex(n < 0 ? 0 : n);
    ui->comboBoxScan->blockSignals(oldState);
}

void ScanSmoothDialog::append()
{
    keepTemporary();
    parseScans();
    compute();
}

void ScanSmoothDialog::keepTemporary()
{
    if (!checkBackend()) return;

    if (temporaryScan) {
        temporaryScan->setTypes(Scan::XY | Scan::SYNTHETIC);
        temporaryScan = nullptr;
        graphControl->updateViews();
    }
}

void ScanSmoothDialog::clearTemporary()
{
    if (!checkBackend()) return;

    if (temporaryScan) {
        QUuid uid = temporaryScan->uid();
        temporaryScan = nullptr;
        graphControl->removeScan(uid);
        graphControl->updateViews();
    }
}

void ScanSmoothDialog::clearGui()
{
    bool oldState = ui->comboBoxScan->blockSignals(true);
    ui->comboBoxScan->clear();
    ui->comboBoxScan->blockSignals(oldState);
}

void ScanSmoothDialog::resetSma()
{
    blockUpdateSignals(true);
    ui->spinBoxSmaWindow->setValue(5);
    computeSma();
    blockUpdateSignals(false);
}

void ScanSmoothDialog::resetEma()
{
    blockUpdateSignals(true);
    ui->spinBoxEmaWindow->setValue(0.5);
    computeEma();
    blockUpdateSignals(false);
}

void ScanSmoothDialog::resetLrma()
{
    blockUpdateSignals(true);
    ui->spinBoxLrmaWindow->setValue(5);
    computeLrma();
    blockUpdateSignals(false);
}

void ScanSmoothDialog::blockUpdateSignals(bool b)
{
    ui->comboBoxScan->blockSignals(b);
    ui->comboBoxAlgo->blockSignals(b);

    ui->spinBoxSmaWindow->blockSignals(b);
    ui->spinBoxEmaWindow->blockSignals(b);
    ui->spinBoxLrmaWindow->blockSignals(b);
}
